import os
from typing import Dict, List, Union

from notdiamond.toolkit._retry import (
    AsyncRetryWrapper,
    ClientType,
    ModelType,
    OpenAIMessagesType,
    RetryManager,
    RetryWrapper,
)


def init(
    client: Union[ClientType, List[ClientType]],
    models: ModelType,
    max_retries: Union[int, Dict[str, int]] = 1,
    timeout: Union[float, Dict[str, float]] = 60.0,
    model_messages: Dict[str, OpenAIMessagesType] = None,
    api_key: Union[str, None] = None,
    async_mode: bool = False,
    backoff: Union[float, Dict[str, float]] = 2.0,
) -> RetryManager:
    """
    Entrypoint for fallback and retry features without changing existing code.

    Add this to existing codebase without other modifications to enable the following capabilities:

    - Fallback to a different model if a model invocation fails.
    - If configured, fallback to a different *provider* if a model invocation fails
      (eg. azure/gpt-4o fails -> invoke openai/gpt-4o)
    - Load-balance between models and providers, if specified.
    - Pass timeout and retry configurations to each invoke, optionally configured per model.
    - Pass model-specific messages on each retry (prepended to the provided `messages` parameter)

    Parameters:
        client (Union[ClientType, List[ClientType]]): Clients to apply retry/fallback logic to.
        models (Union[Dict[str, float], List[str]]):
            Models to use of the format <provider>/<model>.
            Supports two formats:
                - List of models, eg. ["openai/gpt-4o", "azure/gpt-4o"]. Models will be prioritized as listed.
                - Dict of models to weights for load balancing, eg. {"openai/gpt-4o": 0.9, "azure/gpt-4o": 0.1}.
                  If a model invocation fails, the next model is selected by sampling using the *remaining* weights.
        max_retries (Union[int, Dict[str, int]]):
            Maximum number of retries. Can be configured globally or per model.
        timeout (Union[float, Dict[str, float]]):
            Timeout in seconds per model. Can be configured globally or per model.
        model_messages (Dict[str, OpenAIMessagesType]):
            Model-specific messages to prepend to `messages` on each invocation, formatted OpenAI-style. Can be
            configured using any role which is valid as an initial message (eg. "system" or "user", but not "assistant").
        api_key (Optional[str]):
            Not Diamond API key for authentication. Unused for now - will offer logging and metrics in the future.
        async_mode (bool):
            Whether to manage clients as async.
        backoff (Union[float, Dict[str, float]]):
            Backoff factor for exponential backoff per each retry. Can be configured globally or per model.

    Returns:
        RetryManager: Manager object that handles retries and fallbacks. Not required for usage.

    Model Fallback Prioritization
    -----------------------------

    - If models is a list, the fallback model is selected in order after removing the failed model.
      eg. If "openai/gpt-4o" fails for the list:
        - ["openai/gpt-4o", "azure/gpt-4o"], "azure/gpt-4o" will be tried next
        - ["openai/gpt-4o-mini", "openai/gpt-4o", "azure/gpt-4o"], "openai/gpt-4o-mini" will be tried next.
    - If models is a dict, the next model is selected by sampling using the *remaining* weights.
      eg. If "openai/gpt-4o" fails for the dict:
        - {"openai/gpt-4o": 0.9, "azure/gpt-4o": 0.1}, "azure/gpt-4o" will be invoked 100% of the time
        - {"openai/gpt-4o": 0.5, "azure/gpt-4o": 0.25, "openai/gpt-4o-mini": 0.25}, then "azure/gpt-4o" and
          "openai/gpt-4o-mini" can be invoked with 50% probability each.

    Usage
    -----

    Please refer to tests/test_init.py for more examples on how to use notdiamond.init.

    .. code-block:: python

        # ...existing workflow code, including client initialization...
        openai_client = OpenAI(...)
        azure_client = AzureOpenAI(...)

        # Add `notdiamond.init` to the workflow.
        notdiamond.init(
            [openai_client, azure_client],
            models={"openai/gpt-4o": 0.9, "azure/gpt-4o": 0.1},
            max_retries={"openai/gpt-4o": 3, "azure/gpt-4o": 1},
            timeout={"openai/gpt-4o": 10.0, "azure/gpt-4o": 5.0},
            model_messages={
                "openai/gpt-4o": [{"role": "user", "content": "Here is a prompt for OpenAI."}],
                "azure/gpt-4o": [{"role": "user", "content": "Here is a prompt for Azure."}],
            },
            api_key="sk-...",
            backoff=2.0,
        )

        # ...continue existing workflow code...
        response = openai_client.chat.completions.create(
            model="notdiamond",
            messages=[{"role": "user", "content": "Hello!"}]
        )

    """
    api_key = api_key or os.getenv("NOTDIAMOND_API_KEY")

    if async_mode:
        wrapper_cls = AsyncRetryWrapper
    else:
        wrapper_cls = RetryWrapper

    for model in models:
        if len(model.split("/")) != 2:
            raise ValueError(
                f"Model {model} must be in the format <provider>/<model>."
            )

    if not isinstance(client, List):
        client_wrappers = [
            wrapper_cls(
                client=client,
                models=models,
                max_retries=max_retries,
                timeout=timeout,
                model_messages=model_messages,
                api_key=api_key,
                backoff=backoff,
            )
        ]
    else:
        client_wrappers = [
            wrapper_cls(
                client=cc,
                models=models,
                max_retries=max_retries,
                timeout=timeout,
                model_messages=model_messages,
                api_key=api_key,
                backoff=backoff,
            )
            for cc in client
        ]
    retry_manager = RetryManager(models, client_wrappers)

    return retry_manager



import os
from importlib import metadata

from dotenv import load_dotenv

load_dotenv(os.getcwd() + "/.env")

VERSION = metadata.version("notdiamond")

NOTDIAMOND_API_KEY = os.getenv("NOTDIAMOND_API_KEY", default="")
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY", default="")
ANTHROPIC_API_KEY = os.getenv("ANTHROPIC_API_KEY", default="")
GOOGLE_API_KEY = os.getenv("GOOGLE_API_KEY", default="")
COHERE_API_KEY = os.getenv("COHERE_API_KEY", default="")
MISTRAL_API_KEY = os.getenv("MISTRAL_API_KEY", default="")
TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY", default="")
PPLX_API_KEY = os.getenv("PPLX_API_KEY", default="")
REPLICATE_API_KEY = os.getenv("REPLICATE_API_KEY", default="")


NOTDIAMOND_API_URL = os.getenv(
    "NOTDIAMOND_API_URL", "https://api.notdiamond.ai"
)

PROVIDERS = {
    "openai": {
        "models": [
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-0125",
            "gpt-4",
            "gpt-4-0613",
            "gpt-4-turbo",
            "gpt-4-turbo-2024-04-09",
            "gpt-4o",
            "gpt-4o-2024-05-13",
            "gpt-4o-2024-08-06",
            "gpt-4o-mini",
            "gpt-4o-mini-2024-07-18",
            "gpt-4-turbo-preview",
            "gpt-4-0125-preview",
            "gpt-4-1106-preview",
            "gpt-4.1",
            "gpt-4.1-2025-04-14",
            "gpt-4.1-mini",
            "gpt-4.1-mini-2025-04-14",
            "gpt-4.1-nano",
            "gpt-4.1-nano-2025-04-14",
            "o1-preview",
            "o1-preview-2024-09-12",
            "o1-mini",
            "o1-mini-2024-09-12",
            "chatgpt-4o-latest",
        ],
        "api_key": OPENAI_API_KEY,
        "support_tools": [
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-0125",
            "gpt-4",
            "gpt-4-0613",
            "gpt-4-turbo",
            "gpt-4-turbo-2024-04-09",
            "gpt-4o",
            "gpt-4o-2024-05-13",
            "gpt-4o-2024-08-06",
            "gpt-4o-mini",
            "gpt-4o-mini-2024-07-18",
            "gpt-4-turbo-preview",
            "gpt-4-0125-preview",
            "gpt-4-1106-preview",
            "gpt-4.1",
            "gpt-4.1-2025-04-14",
            "gpt-4.1-mini",
            "gpt-4.1-mini-2025-04-14",
            "gpt-4.1-nano",
            "gpt-4.1-nano-2025-04-14",
        ],
        "support_response_model": [
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-0125",
            "gpt-4",
            "gpt-4-0613",
            "gpt-4-turbo",
            "gpt-4-turbo-2024-04-09",
            "gpt-4o",
            "gpt-4o-2024-05-13",
            "gpt-4o-2024-08-06",
            "gpt-4o-mini",
            "gpt-4o-mini-2024-07-18",
            "gpt-4-turbo-preview",
            "gpt-4-0125-preview",
            "gpt-4-1106-preview",
            "o1-preview",
            "o1-preview-2024-09-12",
            "o1-mini",
            "o1-mini-2024-09-12",
            "chatgpt-4o-latest",
            "gpt-4.1",
            "gpt-4.1-2025-04-14",
            "gpt-4.1-mini",
            "gpt-4.1-mini-2025-04-14",
            "gpt-4.1-nano",
            "gpt-4.1-nano-2025-04-14",
        ],
        "openrouter_identifier": {
            "gpt-3.5-turbo": "openai/gpt-3.5-turbo",
            "gpt-4": "openai/gpt-4",
            "gpt-4-turbo": "openai/gpt-4-turbo",
            "gpt-4o": "openai/gpt-4o",
            "gpt-4o-2024-05-13": "openai/gpt-4o-2024-05-13",
            "gpt-4o-2024-08-06": "openai/gpt-4o-2024-08-06",
            "gpt-4o-mini": "openai/gpt-4o-mini",
            "gpt-4o-mini-2024-07-18": "openai/gpt-4o-mini-2024-07-18",
            "gpt-4-turbo-preview": "openai/gpt-4-turbo-preview",
            "gpt-4-1106-preview": "openai/gpt-4-1106-preview",
            "o1-preview": "openai/o1-preview",
            "o1-preview-2024-09-12": "openai/o1-preview-2024-09-12",
            "o1-mini": "openai/o1-mini",
            "o1-mini-2024-09-12": "openai/o1-mini-2024-09-12",
            "chatgpt-4o-latest": "openai/chatgpt-4o-latest",
            "gpt-4.1": "openai/gpt-4.1",
            "gpt-4.1-2025-04-14": "openai/gpt-4.1-2025-04-14",
            "gpt-4.1-mini": "openai/gpt-4.1-mini",
            "gpt-4.1-mini-2025-04-14": "openai/gpt-4.1-mini-2025-04-14",
            "gpt-4.1-nano": "openai/gpt-4.1-nano",
            "gpt-4.1-nano-2025-04-14": "openai/gpt-4.1-nano-2025-04-14",
        },
        "price": {
            "gpt-3.5-turbo": {"input": 0.5, "output": 1.5},
            "gpt-3.5-turbo-0125": {"input": 0.5, "output": 1.5},
            "gpt-4": {"input": 30.0, "output": 60.0},
            "gpt-4-0613": {"input": 30.0, "output": 60.0},
            "gpt-4-turbo": {"input": 10.0, "output": 30.0},
            "gpt-4-turbo-2024-04-09": {"input": 10.0, "output": 30.0},
            "gpt-4o": {"input": 5.0, "output": 15.0},
            "gpt-4o-2024-05-13": {"input": 5.0, "output": 15.0},
            "gpt-4o-2024-08-06": {"input": 2.5, "output": 10.0},
            "gpt-4o-mini": {"input": 0.15, "output": 0.6},
            "gpt-4o-mini-2024-07-18": {"input": 0.15, "output": 0.6},
            "gpt-4-turbo-preview": {"input": 10.0, "output": 30.0},
            "gpt-4-0125-preview": {"input": 10.0, "output": 30.0},
            "gpt-4-1106-preview": {"input": 10.0, "output": 30.0},
            "o1-preview": {"input": 15.0, "output": 60.0},
            "o1-preview-2024-09-12": {"input": 15.0, "output": 60.0},
            "o1-mini": {"input": 3.0, "output": 12.0},
            "o1-mini-2024-09-12": {"input": 3.0, "output": 12.0},
            "chatgpt-4o-latest": {"input": 5.0, "output": 15.0},
            "gpt-4.1": {"input": 2.0, "output": 8.0},
            "gpt-4.1-mini": {"input": 0.5, "output": 1.6},
            "gpt-4.1-nano": {"input": 0.1, "output": 0.4},
            "gpt-4.1-2025-04-14": {"input": 2.0, "output": 8.0},
            "gpt-4.1-mini-2025-04-14": {"input": 0.5, "output": 1.6},
            "gpt-4.1-nano-2025-04-14": {"input": 0.1, "output": 0.4},
        },
    },
    "anthropic": {
        "models": [
            "claude-2.1",
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
            "claude-3-5-haiku-20241022",
            "claude-3-5-sonnet-20240620",
            "claude-3-5-sonnet-20241022",
            "claude-3-5-sonnet-latest",
            "claude-3-7-sonnet-latest",
            "claude-3-7-sonnet-20250219",
            "claude-opus-4-20250514",
            "claude-sonnet-4-20250514",
            "claude-opus-4-0",
            "claude-sonnet-4-0",
        ],
        "api_key": ANTHROPIC_API_KEY,
        "support_tools": [
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
            "claude-3-5-sonnet-20240620",
            "claude-3-5-sonnet-20241022",
            "claude-3-5-haiku-20241022",
            "claude-3-5-sonnet-latest",
            "claude-3-7-sonnet-latest",
            "claude-3-7-sonnet-20250219",
            "claude-opus-4-20250514",
            "claude-sonnet-4-20250514",
            "claude-opus-4-0",
            "claude-sonnet-4-0",
        ],
        "support_response_model": [
            "claude-2.1",
            "claude-3-opus-20240229",
        ],
        "openrouter_identifier": {
            "claude-2.1": "anthropic/claude-2.1",
            "claude-3-opus-20240229": "anthropic/claude-3-opus",
            "claude-3-sonnet-20240229": "anthropic/claude-3-sonnet",
            "claude-3-haiku-20240307": "anthropic/claude-3-haiku",
            "claude-3-5-sonnet-20240620": "anthropic/claude-3.5-sonnet-20240620",
            "claude-3-5-sonnet-latest": "anthropic/claude-3.5-sonnet",
            "claude-3-5-haiku-20241022": "anthropic/claude-3.5-haiku",
            "claude-3-7-sonnet-latest": "anthropic/claude-3.7-sonnet",
            "claude-3-7-sonnet-20250219": "anthropic/claude-3.7-sonnet",
            "claude-opus-4-20250514": "anthropic/claude-opus-4",
            "claude-sonnet-4-20250514": "anthropic/claude-sonnet-4",
            "claude-opus-4-0": "anthropic/claude-opus-4",
            "claude-sonnet-4-0": "anthropic/claude-sonnet-4",
        },
        "price": {
            "claude-2.1": {"input": 8.0, "output": 24.0},
            "claude-3-opus-20240229": {"input": 15.0, "output": 75.0},
            "claude-3-sonnet-20240229": {"input": 3.0, "output": 15.0},
            "claude-3-haiku-20240307": {"input": 0.25, "output": 1.25},
            "claude-3-5-haiku-20241022": {"input": 1.0, "output": 5.0},
            "claude-3-5-sonnet-20240620": {"input": 3.0, "output": 15.0},
            "claude-3-5-sonnet-20241022": {"input": 3.0, "output": 15.0},
            "claude-3-5-sonnet-latest": {"input": 3.0, "output": 15.0},
            "claude-3-7-sonnet-latest": {"input": 3.0, "output": 15.0},
            "claude-3-7-sonnet-20250219": {"input": 3.0, "output": 15.0},
            "claude-opus-4-20250514": {"input": 15.0, "output": 75.0},
            "claude-sonnet-4-20250514": {"input": 3.0, "output": 15.0},
            "claude-opus-4-0": {"input": 15.0, "output": 75.0},
            "claude-sonnet-4-0": {"input": 3.0, "output": 15.0},
        },
    },
    "google": {
        "models": [
            "gemini-pro",
            "gemini-1.0-pro-latest",
            "gemini-1.5-pro-latest",
            "gemini-1.5-pro-exp-0801",
            "gemini-1.5-flash-latest",
            "gemini-2.0-flash",
            "gemini-2.0-flash-001",
            "gemini-2.5-flash",
            "gemini-2.5-pro",
        ],
        "api_key": GOOGLE_API_KEY,
        "support_tools": [
            "gemini-pro",
            "gemini-1.0-pro-latest",
            "gemini-1.5-pro-latest",
            "gemini-1.5-pro-exp-0801",
            "gemini-1.5-flash-latest",
            "gemini-2.5-flash",
            "gemini-2.5-pro",
        ],
        "support_response_model": [
            "gemini-pro",
            "gemini-1.0-pro-latest",
            "gemini-1.5-pro-latest",
            "gemini-1.5-pro-exp-0801",
            "gemini-1.5-flash-latest",
            "gemini-2.0-flash",
            "gemini-2.0-flash-001",
            "gemini-2.5-flash",
            "gemini-2.5-pro",
        ],
        "openrouter_identifier": {
            "gemini-pro": "google/gemini-pro",
            "gemini-1.0-pro-latest": "google/gemini-pro",
            "gemini-1.5-pro-latest": "google/gemini-pro-1.5",  #
            "gemini-1.5-pro-exp-0801": "google/gemini-pro-1.5-exp",  #
            "gemini-1.5-flash-latest": "google/gemini-flash-1.5",  #
            "gemini-2.0-flash": "google/gemini-2.0-flash",  #
            "gemini-2.0-flash-001": "google/gemini-2.0-flash",  #
            "gemini-2.5-flash": "google/gemini-2.5-flash",  #
            "gemini-2.5-pro": "google/gemini-2.5-pro",  #
        },
        "price": {
            "gemini-pro": {"input": 0.5, "output": 1.5},
            "gemini-1.0-pro-latest": {"input": 0.5, "output": 1.5},
            "gemini-1.5-pro-latest": {"input": 1.75, "output": 10.5},
            "gemini-1.5-pro-exp-0801": {"input": 1.75, "output": 10.5},
            "gemini-1.5-flash-latest": {"input": 0.35, "output": 1.05},
            "gemini-2.0-flash": {"input": 0.10, "output": 0.40},
            "gemini-2.0-flash-001": {"input": 0.10, "output": 0.40},
            "gemini-2.5-flash": {"input": 0.15, "output": 0.6},
            "gemini-2.5-pro": {"input": 2.0, "output": 10.0},
        },
    },
    "cohere": {
        "models": ["command-r", "command-r-plus"],
        "api_key": COHERE_API_KEY,
        "support_tools": ["command-r", "command-r-plus"],
        "support_response_model": ["command-r", "command-r-plus"],
        "openrouter_identifier": {
            "command-r": "cohere/command-r",
            "command-r-plus": "cohere/command-r-plus",
        },
        "price": {
            "command-r": {"input": 0.5, "output": 1.5},
            "command-r-plus": {"input": 3.0, "output": 15.0},
        },
    },
    "mistral": {
        "models": [
            "mistral-large-latest",
            "mistral-large-2402",
            "mistral-large-2407",
            "mistral-medium-latest",
            "mistral-small-latest",
            "open-mistral-7b",
            "open-mixtral-8x7b",
            "open-mixtral-8x22b",
            "open-mistral-nemo",
            "codestral-latest",
        ],
        "api_key": MISTRAL_API_KEY,
        "support_tools": [
            "mistral-large-latest",
            "mistral-large-2402",
            "mistral-large-2407",
            "mistral-small-latest",
            "open-mixtral-8x22b",
            "open-mistral-nemo",
        ],
        "support_response_model": [
            "mistral-large-latest",
            "mistral-large-2402",
            "mistral-large-2407",
            "mistral-medium-latest",
            "mistral-small-latest",
            "open-mistral-7b",
            "open-mixtral-8x7b",
            "open-mixtral-8x22b",
            "open-mistral-nemo",
            "codestral-latest",
        ],
        "openrouter_identifier": {
            "mistral-large-latest": "mistralai/mistral-large",
            "mistral-large-2407": "mistralai/mistral-large",
            "mistral-medium-latest": "mistralai/mistral-medium",
            "mistral-small-latest": "mistralai/mistral-small",
            "open-mistral-7b": "mistralai/mistral-7b-instruct",
            "open-mixtral-8x7b": "mistralai/mixtral-8x7b",
            "open-mixtral-8x22b": "mistralai/mixtral-8x22b-instruct",
            "open-mistral-nemo": "mistralai/mistral-nemo",
        },
        "price": {
            "mistral-large-latest": {"input": 2.0, "output": 6.0},
            "mistral-large-2402": {"input": 4.0, "output": 12.0},
            "mistral-large-2407": {"input": 2.0, "output": 6.0},
            "mistral-medium-latest": {"input": 2.7, "output": 8.1},
            "mistral-small-latest": {"input": 1.0, "output": 3.0},
            "open-mistral-7b": {"input": 0.25, "output": 0.25},
            "open-mixtral-8x7b": {"input": 0.7, "output": 0.7},
            "open-mixtral-8x22b": {"input": 2.0, "output": 6.0},
            "open-mistral-nemo": {"input": 0.15, "output": 0.15},
            "codestral-latest": {"input": 1.0, "output": 3.0},
        },
    },
    "togetherai": {
        "models": [
            "Mistral-7B-Instruct-v0.2",
            "Mixtral-8x7B-Instruct-v0.1",
            "Mixtral-8x22B-Instruct-v0.1",
            "Llama-3-70b-chat-hf",
            "Llama-3-8b-chat-hf",
            "Qwen2-72B-Instruct",
            "Meta-Llama-3.1-8B-Instruct-Turbo",
            "Meta-Llama-3.1-70B-Instruct-Turbo",
            "Meta-Llama-3.1-405B-Instruct-Turbo",
            "DeepSeek-R1",
        ],
        "api_key": TOGETHER_API_KEY,
        "model_prefix": {
            "Mistral-7B-Instruct-v0.2": "mistralai",
            "Mixtral-8x7B-Instruct-v0.1": "mistralai",
            "Mixtral-8x22B-Instruct-v0.1": "mistralai",
            "Llama-3-70b-chat-hf": "meta-llama",
            "Llama-3-8b-chat-hf": "meta-llama",
            "Qwen2-72B-Instruct": "Qwen",
            "Meta-Llama-3.1-8B-Instruct-Turbo": "meta-llama",
            "Meta-Llama-3.1-70B-Instruct-Turbo": "meta-llama",
            "Meta-Llama-3.1-405B-Instruct-Turbo": "meta-llama",
            "DeepSeek-R1": "deepseek-ai",
        },
        "openrouter_identifier": {
            "Llama-3-70b-chat-hf": "meta-llama/llama-3-70b-instruct",
            "Llama-3-8b-chat-hf": "meta-llama/llama-3-8b-instruct",
            "Qwen2-72B-Instruct": "qwen/qwen-2-72b-instruct",
            "Mistral-7B-Instruct-v0.2": "mistralai/mistral-7b-instruct",
            "Mixtral-8x7B-Instruct-v0.1": "mistralai/mixtral-8x7b",
            "Mixtral-8x22B-Instruct-v0.1": "mistralai/mixtral-8x22b-instruct",
            "Meta-Llama-3.1-8B-Instruct-Turbo": "meta-llama/llama-3.1-8b-instruct",
            "Meta-Llama-3.1-70B-Instruct-Turbo": "meta-llama/llama-3.1-70b-instruct",
            "Meta-Llama-3.1-405B-Instruct-Turbo": "meta-llama/llama-3.1-405b-instruct",
            "DeepSeek-R1": "deepseek/deepseek-r1",
        },
        "price": {
            "Mistral-7B-Instruct-v0.2": {"input": 0.2, "output": 0.2},
            "Mixtral-8x7B-Instruct-v0.1": {"input": 0.6, "output": 0.6},
            "Mixtral-8x22B-Instruct-v0.1": {"input": 0.6, "output": 0.6},
            "Llama-3-70b-chat-hf": {"input": 0.7, "output": 0.7},
            "Llama-3-8b-chat-hf": {"input": 0.2, "output": 0.2},
            "Qwen2-72B-Instruct": {"input": 0.9, "output": 0.9},
            "Meta-Llama-3.1-8B-Instruct-Turbo": {
                "input": 0.18,
                "output": 0.18,
            },
            "Meta-Llama-3.1-70B-Instruct-Turbo": {
                "input": 0.88,
                "output": 0.88,
            },
            "Meta-Llama-3.1-405B-Instruct-Turbo": {
                "input": 5.0,
                "output": 15.0,
            },
            "DeepSeek-R1": {
                "input": 7.0,
                "output": 7.0,
            },
        },
    },
    "perplexity": {
        "models": [
            "sonar",
        ],
        "api_key": PPLX_API_KEY,
        "openrouter_identifier": {
            "sonar": "perplexity/sonar",
        },
        "price": {
            "sonar": {"input": 1.0, "output": 1.0},
        },
    },
    "replicate": {
        "models": [
            "mistral-7b-instruct-v0.2",
            "mixtral-8x7b-instruct-v0.1",
            "meta-llama-3-70b-instruct",
            "meta-llama-3-8b-instruct",
            "meta-llama-3.1-405b-instruct",
        ],
        "api_key": REPLICATE_API_KEY,
        "model_prefix": {
            "mistral-7b-instruct-v0.2": "mistralai",
            "mixtral-8x7b-instruct-v0.1": "mistralai",
            "meta-llama-3-70b-instruct": "meta",
            "meta-llama-3-8b-instruct": "meta",
            "meta-llama-3.1-405b-instruct": "meta",
        },
        "openrouter_identifier": {
            "mistral-7b-instruct-v0.2": "mistralai/mistral-7b-instruct",
            "mixtral-8x7b-instruct-v0.1": "mistralai/mixtral-8x7b",
            "meta-llama-3-70b-instruct": "meta-llama/llama-3-70b-instruct",
            "meta-llama-3-8b-instruct": "meta-llama/llama-3-8b-instruct",
            "meta-llama-3.1-405b-instruct": "meta-llama/llama-3.1-405b-instruct",
        },
        "price": {
            "mistral-7b-instruct-v0.2": {"input": 0.2, "output": 0.2},
            "mixtral-8x7b-instruct-v0.1": {"input": 0.6, "output": 0.6},
            "meta-llama-3-70b-instruct": {"input": 0.65, "output": 2.75},
            "meta-llama-3-8b-instruct": {"input": 0.05, "output": 0.25},
            "meta-llama-3.1-405b-instruct": {"input": 9.5, "output": 9.5},
        },
    },
}


EMBEDDING_PROVIDERS = {
    "openai": {
        "models": [
            "text-embedding-3-large",
            "text-embedding-3-small",
            "text-embedding-ada-002",
        ],
        "api_key": OPENAI_API_KEY,
    },
    "cohere": {
        "models": [
            "embed-english-v3.0",
            "embed-english-light-v3.0",
            "embed-multilingual-v3.0",
            "embed-multilingual-light-v3.0",
        ],
        "api_key": COHERE_API_KEY,
    },
    "mistral": {
        "models": [
            "mistral-embed",
        ],
        "api_key": MISTRAL_API_KEY,
    },
}

DEFAULT_USER_AGENT = f"Python-SDK/{VERSION}"



import json
import tempfile
import time
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import requests
from tqdm import tqdm

from notdiamond.exceptions import ApiError
from notdiamond.llms.client import NotDiamond
from notdiamond.llms.config import LLMConfig
from notdiamond.settings import NOTDIAMOND_API_KEY, NOTDIAMOND_API_URL, VERSION
from notdiamond.types import NDApiKeyValidator
from notdiamond._utils import token_counter


class CustomRouter:
    """
    Implementation of CustomRouter class, used to train custom routers using custom datasets.

    Attributes:
        language (str): The language of the dataset in lowercase. Defaults to "english".
        maximize (bool): Whether higher score is better. Defaults to true.
        api_key (Optional[str], optional): The NotDiamond API key. If not specified, will try to
            find it in the environment variable NOTDIAMOND_API_KEY.
    """

    def __init__(
        self,
        language: str = "english",
        maximize: bool = True,
        api_key: Optional[str] = None,
    ):
        if api_key is None:
            api_key = NOTDIAMOND_API_KEY
        NDApiKeyValidator(api_key=api_key)

        self.api_key = api_key
        self.language = language
        self.maximize = maximize

    def _request_train_router(
        self,
        prompt_column: str,
        dataset_file: str,
        llm_configs: List[LLMConfig],
        preference_id: Optional[str],
        nd_api_url: str,
    ) -> str:
        url = f"{nd_api_url}/v2/pzn/trainCustomRouter"

        files = {"dataset_file": open(dataset_file, "rb")}

        payload = {
            "language": self.language,
            "llm_providers": json.dumps(
                [provider.prepare_for_request() for provider in llm_configs]
            ),
            "prompt_column": prompt_column,
            "maximize": self.maximize,
            "preference_id": preference_id,
        }

        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "User-Agent": f"Python-SDK/{VERSION}",
        }

        response = requests.post(
            url=url, headers=headers, data=payload, files=files
        )
        if response.status_code != 200:
            raise ApiError(
                f"ND backend error status code: {response.status_code}, {response.text}"
            )

        preference_id = response.json()["preference_id"]
        return preference_id

    def _prepare_joint_dataset(
        self,
        dataset: Dict[Union[str, LLMConfig], pd.DataFrame],
        prompt_column: str,
        response_column: str,
        score_column: str,
    ) -> Tuple[pd.DataFrame, List[LLMConfig]]:
        a_provider = list(dataset.keys())[0]
        prompts = dataset[a_provider].get(prompt_column, None)
        if prompts is None:
            raise ValueError(f"Prompt column {prompt_column} not found in df.")
        prompts = prompts.to_list()

        llm_configs = []
        joint_dataset = {prompt_column: prompts}
        for provider, df in dataset.items():
            llm_configs.append(provider)

            responses = df.get(response_column, None)
            if responses is None:
                raise ValueError(
                    f"Response column {response_column} not found in df."
                )
            responses = responses.to_list()
            joint_dataset[f"{str(provider)}/response"] = responses

            scores = df.get(score_column, None)
            if scores is None:
                raise ValueError(
                    f"Score column {score_column} not found in df."
                )
            scores = scores.to_list()
            joint_dataset[f"{str(provider)}/score"] = scores

        joint_df = pd.DataFrame(joint_dataset)

        llm_configs = NotDiamond._parse_llm_configs_data(llm_configs)
        return joint_df, llm_configs

    def fit(
        self,
        dataset: Dict[Union[str, LLMConfig], pd.DataFrame],
        prompt_column: str,
        response_column: str,
        score_column: str,
        preference_id: Optional[str] = None,
        nd_api_url: Optional[str] = NOTDIAMOND_API_URL,
    ) -> str:
        """
        Method to train a custom router using provided dataset.

        Parameters:
            dataset (Dict[str, pandas.DataFrame]): The dataset to train a custom router.
                Each key in the dictionary should be in the form of <provider>/<model>.
            prompt_column (str): The column name in each DataFrame corresponding
                to the prompts used to evaluate the LLM.
            response_column (str): The column name in each DataFrame corresponding
                to the response given by the LLM for a given prompt.
            score_column (str): The column name in each DataFrame corresponding
                to the score given to the response from the LLM.
            preference_id (Optional[str], optional): If specified, the custom router
                associated with the preference_id will be updated with the provided dataset.
            nd_api_url (Optional[str], optional): The URL of the NotDiamond API. Defaults to prod.

        Raises:
            ApiError: When the NotDiamond API fails
            ValueError: When parsing the provided dataset fails
            UnsupportedLLMProvider: When a provider specified in the dataset is not supported.

        Returns:
            str:
                preference_id: the preference_id associated with the custom router.
                    Use this preference_id in your routing calls to use the custom router.
        """

        joint_df, llm_configs = self._prepare_joint_dataset(
            dataset, prompt_column, response_column, score_column
        )

        with tempfile.NamedTemporaryFile(suffix=".csv") as joint_csv:
            joint_df.to_csv(joint_csv.name, index=False)
            preference_id = self._request_train_router(
                prompt_column,
                joint_csv.name,
                llm_configs,
                preference_id,
                nd_api_url,
            )

        return preference_id

    def _get_latency(self, llm_config: LLMConfig, prompt: str) -> float:
        llm = NotDiamond._llm_from_config(llm_config)
        start_time = time.time()
        _ = llm.invoke([("human", prompt)])
        end_time = time.time()
        return (end_time - start_time) * 1000  # ms

    def _get_cost(
        self, llm_config: LLMConfig, prompt: str, response: str
    ) -> float:
        n_input_tokens = token_counter(model="gpt-4o", text=prompt)
        n_output_tokens = token_counter(model="gpt-4o", text=response)
        input_price = (
            llm_config.default_input_price
            if llm_config.input_price is None
            else llm_config.input_price
        )
        output_price = (
            llm_config.default_output_price
            if llm_config.output_price is None
            else llm_config.output_price
        )
        return (
            n_input_tokens * input_price + n_output_tokens * output_price
        ) / 1e6

    def _eval_custom_router(
        self,
        client: NotDiamond,
        llm_configs: List[LLMConfig],
        joint_df: pd.DataFrame,
        prompt_column: str,
        include_latency: bool,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        eval_results = OrderedDict()
        eval_results[prompt_column] = []
        eval_results["session_id"] = []
        eval_results["notdiamond/score"] = []
        eval_results["notdiamond/cost"] = []
        eval_results["notdiamond/response"] = []
        eval_results["notdiamond/recommended_provider"] = []

        if include_latency:
            eval_results["notdiamond/latency"] = []

        for provider in llm_configs:
            provider_score_column = (
                f"{provider.provider}/{provider.model}/score"
            )
            eval_results[provider_score_column] = []

            provider_response_column = (
                f"{provider.provider}/{provider.model}/response"
            )
            eval_results[provider_response_column] = []

            provider_cost_column = f"{provider.provider}/{provider.model}/cost"
            eval_results[provider_cost_column] = []

            if include_latency:
                provider_latency_column = (
                    f"{provider.provider}/{provider.model}/latency"
                )
                eval_results[provider_latency_column] = []

        for _, row in tqdm(joint_df.iterrows(), total=len(joint_df)):
            prompt = row[prompt_column]
            eval_results[prompt_column].append(prompt)

            session_id, nd_provider = client.chat.completions.model_select(
                messages=[{"role": "user", "content": prompt}], timeout=60
            )
            if nd_provider is None:
                continue

            eval_results["session_id"].append(session_id)

            provider_matched = False
            for provider in llm_configs:
                provider_score = row[
                    f"{provider.provider}/{provider.model}/score"
                ]
                eval_results[
                    f"{provider.provider}/{provider.model}/score"
                ].append(provider_score)

                provider_response = row[
                    f"{provider.provider}/{provider.model}/response"
                ]
                eval_results[
                    f"{provider.provider}/{provider.model}/response"
                ].append(provider_response)

                provider_cost = self._get_cost(
                    provider, prompt, provider_response
                )
                eval_results[
                    f"{provider.provider}/{provider.model}/cost"
                ].append(provider_cost)

                if include_latency:
                    provider_latency = self._get_latency(provider, prompt)
                    eval_results[
                        f"{provider.provider}/{provider.model}/latency"
                    ].append(provider_latency)

                if (
                    not provider_matched
                    and provider.provider == nd_provider.provider
                    and provider.model == nd_provider.model
                ):
                    provider_matched = True
                    eval_results["notdiamond/score"].append(provider_score)
                    eval_results["notdiamond/cost"].append(provider_cost)
                    eval_results["notdiamond/response"].append(
                        provider_response
                    )
                    eval_results["notdiamond/recommended_provider"].append(
                        f"{nd_provider.provider}/{nd_provider.model}"
                    )
                    if include_latency:
                        eval_results["notdiamond/latency"].append(
                            provider_latency
                        )

            if not provider_matched:
                raise ValueError(
                    f"""
                    Custom router returned {nd_provider.provider}/{nd_provider.model}
                    which is not in the set of models in the test dataset
                    """
                )

        eval_results_df = pd.DataFrame(eval_results)

        eval_stats = OrderedDict()
        best_average_provider = None
        best_average_score = -(2 * int(self.maximize) - 1) * np.inf

        nd_average_score = eval_results_df["notdiamond/score"].mean()
        eval_stats["Not Diamond Average Score"] = [nd_average_score]

        nd_average_cost = eval_results_df["notdiamond/cost"].mean()
        eval_stats["Not Diamond Average Cost"] = [nd_average_cost]

        if include_latency:
            nd_average_latency = eval_results_df["notdiamond/latency"].mean()
            eval_stats["Not Diamond Average Latency"] = [nd_average_latency]

        for provider in llm_configs:
            provider_avg_score = eval_results_df[
                f"{provider.provider}/{provider.model}/score"
            ].mean()
            eval_stats[f"{provider.provider}/{provider.model}/avg_score"] = [
                provider_avg_score
            ]

            provider_avg_cost = eval_results_df[
                f"{provider.provider}/{provider.model}/cost"
            ].mean()
            eval_stats[f"{provider.provider}/{provider.model}/avg_cost"] = [
                provider_avg_cost
            ]

            if include_latency:
                provider_avg_latency = eval_results_df[
                    f"{provider.provider}/{provider.model}/latency"
                ].mean()
                eval_stats[
                    f"{provider.provider}/{provider.model}/avg_latency"
                ] = [provider_avg_latency]

            if self.maximize:
                if provider_avg_score > best_average_score:
                    best_average_score = provider_avg_score
                    best_average_cost = provider_avg_cost
                    best_average_provider = (
                        f"{provider.provider}/{provider.model}"
                    )
                    if include_latency:
                        best_average_latency = provider_avg_latency
            else:
                if provider_avg_score < best_average_score:
                    best_average_score = provider_avg_score
                    best_average_cost = provider_avg_cost
                    best_average_provider = (
                        f"{provider.provider}/{provider.model}"
                    )
                    if include_latency:
                        best_average_latency = provider_avg_latency

        eval_stats["Best Average Provider"] = [best_average_provider]
        eval_stats["Best Provider Average Score"] = [best_average_score]
        eval_stats["Best Provider Average Cost"] = [best_average_cost]

        if include_latency:
            eval_stats["Best Provider Average Latency"] = [
                best_average_latency
            ]

            first_columns = [
                "Best Average Provider",
                "Best Provider Average Score",
                "Best Provider Average Cost",
                "Best Provider Average Latency",
                "Not Diamond Average Score",
                "Not Diamond Average Cost",
                "Not Diamond Average Latency",
            ]
        else:
            first_columns = [
                "Best Average Provider",
                "Best Provider Average Score",
                "Best Provider Average Cost",
                "Not Diamond Average Score",
                "Not Diamond Average Cost",
            ]

        column_order = first_columns + [
            col for col in eval_stats.keys() if col not in first_columns
        ]
        ordered_eval_stats = OrderedDict()
        for col in column_order:
            ordered_eval_stats[col] = eval_stats[col]

        eval_stats_df = pd.DataFrame(ordered_eval_stats)
        return eval_results_df, eval_stats_df

    def eval(
        self,
        dataset: Dict[Union[str, LLMConfig], pd.DataFrame],
        prompt_column: str,
        response_column: str,
        score_column: str,
        preference_id: str,
        include_latency: bool = False,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Method to evaluate a custom router using provided dataset.

        Parameters:
            dataset (Dict[str, pandas.DataFrame]): The dataset to train a custom router.
                Each key in the dictionary should be in the form of <provider>/<model>.
            prompt_column (str): The column name in each DataFrame corresponding
                to the prompts used to evaluate the LLM.
            response_column (str): The column name in each DataFrame corresponding
                to the response given by the LLM for a given prompt.
            score_column (str): The column name in each DataFrame corresponding
                to the score given to the response from the LLM.
            preference_id (str): The preference_id associated with the custom router
                returned from .fit().

        Raises:
            ApiError: When the NotDiamond API fails
            ValueError: When parsing the provided dataset fails
            UnsupportedLLMProvider: When a provider specified in the dataset is not supported.

        Returns:
            Tuple[pandas.DataFrame, pandas.DataFrame]:
                eval_results_df: A DataFrame containing all the prompts, responses of each provider
                    (indicated by column <provider>/<model>/response), scores of each provider
                    (indicated by column <provider>/<model>/score), and notdiamond custom router
                    response and score (indicated by column notdiamond/response and notdiamond/score).
                eval_stats_df: A DataFrame containing the "Best Average Provider" computed from the
                    provided dataset, the "Best Provider Average Score" achieved by the "Best Average Provider",
                    and the "Not Diamond Average Score" achieved through custom router.
        """

        joint_df, llm_configs = self._prepare_joint_dataset(
            dataset, prompt_column, response_column, score_column
        )

        client = NotDiamond(
            llm_configs=llm_configs,
            api_key=self.api_key,
            preference_id=preference_id,
        )

        eval_results_df, eval_stats_df = self._eval_custom_router(
            client, llm_configs, joint_df, prompt_column, include_latency
        )
        return eval_results_df, eval_stats_df



from typing import Union

from langchain_cohere import CohereEmbeddings
from langchain_mistralai import MistralAIEmbeddings
from langchain_openai import OpenAIEmbeddings
from ragas.embeddings import HuggingfaceEmbeddings, LangchainEmbeddingsWrapper
from ragas.llms import LangchainLLMWrapper

from ...exceptions import UnsupportedEmbeddingProvider
from ...llms.client import NotDiamond
from ...llms.config import EmbeddingConfig, LLMConfig


def get_llm(llm_config_or_str: Union[LLMConfig, str]) -> LangchainLLMWrapper:
    """
    Build the LLM object compatible with evaluation metrics.

    Parameters:
        llm_config_or_str (Union[LLMConfig, str]): a LLMConfig object or a model string
            that specifies the LLM to construct.
    """
    if isinstance(llm_config_or_str, str):
        llm_config = LLMConfig.from_string(llm_config_or_str)
    else:
        llm_config = llm_config_or_str

    lc_llm = NotDiamond._llm_from_config(llm_config)
    return LangchainLLMWrapper(lc_llm)


def get_embedding(
    embedding_model_config_or_str: Union[EmbeddingConfig, str]
) -> Union[LangchainEmbeddingsWrapper, HuggingfaceEmbeddings]:
    """
    Build the embedding model object compatible with evaluation metrics.

    Parameters:
        embedding_model_config_or_str (Union[EmbeddingConfig, str]): an EmbeddingConfig object
            or an embedding model string that specifies the embedding model to construct.
    """
    if isinstance(embedding_model_config_or_str, str):
        embedding_config = EmbeddingConfig.from_string(
            embedding_model_config_or_str
        )
    else:
        embedding_config = embedding_model_config_or_str

    if embedding_config.provider == "openai":
        lc_embedding = OpenAIEmbeddings(
            model=embedding_config.model, **embedding_config.kwargs
        )

    elif embedding_config.provider == "cohere":
        lc_embedding = CohereEmbeddings(
            model=embedding_config.model, **embedding_config.kwargs
        )

    elif embedding_config.provider == "mistral":
        lc_embedding = MistralAIEmbeddings(
            model=embedding_config.model, **embedding_config.kwargs
        )

    elif embedding_config.provider == "huggingface":
        return HuggingfaceEmbeddings(model_name=embedding_config.model)

    else:
        raise UnsupportedEmbeddingProvider(
            f"Embedding model {str(embedding_config)} not supported."
        )
    return LangchainEmbeddingsWrapper(lc_embedding)



from typing import Optional

from notdiamond import settings
from notdiamond.exceptions import ApiError
from notdiamond.llms.config import LLMConfig
from notdiamond.metrics.request import feedback_request
from notdiamond.types import NDApiKeyValidator


class Metric:
    def __init__(self, metric: Optional[str] = "accuracy"):
        self.metric = metric

    def __call__(self):
        return self.metric

    def feedback(
        self,
        session_id: str,
        llm_config: LLMConfig,
        value: int,
        notdiamond_api_key: Optional[str] = None,
        _user_agent: str = None,
    ):
        if notdiamond_api_key is None:
            notdiamond_api_key = settings.NOTDIAMOND_API_KEY
        NDApiKeyValidator(api_key=notdiamond_api_key)
        if value not in [0, 1]:
            raise ApiError("Invalid feedback value. It must be 0 or 1.")

        return feedback_request(
            session_id=session_id,
            llm_config=llm_config,
            feedback_payload=self.request_payload(value),
            notdiamond_api_key=notdiamond_api_key,
            _user_agent=_user_agent,
        )

    def request_payload(self, value: int):
        return {self.metric: value}



"""NotDiamond client class"""


import inspect
import logging
import time
import warnings
from enum import Enum
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

# Details: https://python.langchain.com/v0.1/docs/guides/development/pydantic_compatibility/
from pydantic import BaseModel
from pydantic_partial import create_partial_model

from notdiamond import settings
from notdiamond._utils import _module_check, token_counter
from notdiamond.exceptions import (
    ApiError,
    CreateUnavailableError,
    MissingLLMConfigs,
)
from notdiamond.llms.config import LLMConfig
from notdiamond.llms.providers import is_o1_model
from notdiamond.llms.request import (
    amodel_select,
    create_preference_id,
    model_select,
    report_latency,
)
from notdiamond.metrics.metric import Metric
from notdiamond.prompts import (
    _curly_escape,
    inject_system_prompt,
    o1_system_prompt_translate,
)
from notdiamond.types import NDApiKeyValidator

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


class _NDClientTarget(Enum):
    ROUTER = "router"
    INVOKER = "invoker"


def _ndllm_factory(import_target: _NDClientTarget = None):
    _invoke_error_msg_tmpl = (
        "{fn_name} is not available. `notdiamond` can generate LLM responses after "
        "installing additional dependencies via `pip install notdiamond[create]`."
    )

    _default_llm_config_invalid_warning = "The default LLMConfig set is invalid. Defaulting to {provider}/{model}"

    _no_default_llm_config_warning = (
        "No default LLMConfig set. Defaulting to {provider}/{model}"
    )

    class _NDRouterClient(BaseModel):
        api_key: str
        llm_configs: Optional[List[Union[LLMConfig, str]]]
        default: Union[LLMConfig, int, str]
        max_model_depth: Optional[int]
        latency_tracking: bool
        hash_content: bool
        tradeoff: Optional[str]
        preference_id: Optional[str]
        tools: Optional[Sequence[Union[Dict[str, Any], Callable]]]
        callbacks: Optional[List]
        nd_api_url: Optional[str]
        user_agent: Union[str, None]
        max_retries: Optional[int]
        timeout: Optional[Union[float, int]]

        class Config:
            arbitrary_types_allowed = True

        def __init__(
            self,
            llm_configs: Optional[List[Union[LLMConfig, str]]] = None,
            api_key: Optional[str] = None,
            default: Union[LLMConfig, int, str] = 0,
            max_model_depth: Optional[int] = None,
            latency_tracking: bool = True,
            hash_content: bool = False,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            callbacks: Optional[List] = None,
            tools: Optional[Sequence[Union[Dict[str, Any], Callable]]] = None,
            nd_api_url: Optional[str] = settings.NOTDIAMOND_API_URL,
            user_agent: Union[str, None] = None,
            max_retries: Optional[int] = 3,
            timeout: Optional[Union[float, int]] = 60.0,
            **kwargs,
        ):
            if api_key is None:
                api_key = settings.NOTDIAMOND_API_KEY
            NDApiKeyValidator(api_key=api_key)

            if user_agent is None:
                user_agent = settings.DEFAULT_USER_AGENT

            if llm_configs is not None:
                llm_configs = self._parse_llm_configs_data(llm_configs)

                if max_model_depth is None:
                    max_model_depth = len(llm_configs)

                if max_model_depth > len(llm_configs):
                    LOGGER.warning(
                        "WARNING: max_model_depth cannot be bigger than the number of LLMs."
                    )
                    max_model_depth = len(llm_configs)

            if tradeoff is not None:
                if tradeoff not in ["cost", "latency"]:
                    raise ValueError(
                        "Invalid tradeoff. Accepted values: cost, latency."
                    )

            if tradeoff is not None:
                warnings.warn(
                    "The tradeoff constructor parameter is deprecated and will be removed in a "
                    "future version. Please specify the tradeoff when using model_select or invocation methods.",
                    DeprecationWarning,
                    stacklevel=2,
                )

            super().__init__(
                api_key=api_key,
                llm_configs=llm_configs,
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
                tools=tools,
                callbacks=callbacks,
                nd_api_url=nd_api_url,
                user_agent=user_agent,
                max_retries=max_retries,
                timeout=timeout,
                **kwargs,
            )
            self.user_agent = user_agent
            assert (
                self.api_key is not None
            ), "API key is not set. Please set a Not Diamond API key."

        @property
        def chat(self):
            return self

        @property
        def completions(self):
            return self

        def create_preference_id(self, name: Optional[str] = None) -> str:
            return create_preference_id(self.api_key, name, self.nd_api_url)

        async def amodel_select(
            self,
            messages: List[Dict[str, str]],
            input: Optional[Dict[str, Any]] = None,
            model: Optional[List[LLMConfig]] = None,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            metric: Metric = Metric("accuracy"),
            previous_session: Optional[str] = None,
            timeout: Optional[Union[float, int]] = None,
            max_retries: Optional[int] = None,
            **kwargs,
        ) -> tuple[str, Optional[LLMConfig]]:
            """
            This function calls the NotDiamond backend to fetch the most suitable model for the given prompt,
            and leaves the execution of the LLM call to the developer.
            The function is async, so it's suitable for async codebases.

            Parameters:
                messages (List[Dict[str, str]]): List of messages, OpenAI style.
                input (Optional[Dict[str, Any]], optional): If the prompt_template contains variables, use input to specify
                                                            the values for those variables. Defaults to None, assuming no
                                                            variables.
                model (Optional[List[LLMConfig]]): List of models to choose from.
                default (Optional[Union[LLMConfig, int, str]]): Default LLM.
                max_model_depth (Optional[int]): If your top recommended model is down, specify up to which depth
                                                    of routing you're willing to go.
                latency_tracking (Optional[bool]): Latency tracking flag.
                hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
                tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                                    for the router to determine the best LLM for a given query.
                preference_id (Optional[str]): The ID of the router preference that was configured via the Dashboard.
                                                Defaults to None.
                metric (Metric, optional): Metric used by NotDiamond router to choose the best LLM.
                                                Defaults to Metric("accuracy").
                previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
                timeout (int): The number of seconds to wait before terminating the API call to Not Diamond backend.
                max_retries (int): The number of retries to attempt before giving up.
                nd_api_url (Optional[str]): The URL of the NotDiamond API. Defaults to settings.NOTDIAMOND_API_URL.
                **kwargs: Any other arguments that are supported by Langchain's invoke method, will be passed through.
            Returns:
                tuple[str, Optional[LLMConfig]]: returns the session_id and the chosen LLM
            """
            if input is None:
                input = {}

            if model is not None:
                llm_configs = self._parse_llm_configs_data(model)
                self.llm_configs = llm_configs

            self.validate_params(
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
            )

            best_llm, session_id = await amodel_select(
                messages=messages,
                llm_configs=self.llm_configs,
                metric=metric,
                notdiamond_api_key=self.api_key,
                max_model_depth=self.max_model_depth,
                hash_content=self.hash_content,
                tradeoff=self.tradeoff,
                preference_id=self.preference_id,
                tools=self.tools,
                previous_session=previous_session,
                timeout=timeout or self.timeout,
                max_retries=max_retries or self.max_retries,
                nd_api_url=self.nd_api_url,
                _user_agent=self.user_agent,
            )

            if not best_llm:
                LOGGER.warning(
                    f"ND API error. Falling back to default provider={self.default_llm.provider}/{self.default_llm.model}."
                )
                best_llm = self.default_llm
            self.call_callbacks("on_model_select", best_llm, best_llm.model)

            return session_id, best_llm

        def model_select(
            self,
            messages: List[Dict[str, str]],
            input: Optional[Dict[str, Any]] = None,
            model: Optional[List[LLMConfig]] = None,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            metric: Metric = Metric("accuracy"),
            previous_session: Optional[str] = None,
            timeout: Optional[Union[float, int]] = None,
            max_retries: Optional[int] = None,
            **kwargs,
        ) -> tuple[str, Optional[LLMConfig]]:
            """
            This function calls the NotDiamond backend to fetch the most suitable model for the given prompt,
            and leaves the execution of the LLM call to the developer.

            Parameters:
                messages (List[Dict[str, str]]): List of messages OpenAI style.
                input (Optional[Dict[str, Any]], optional): If the prompt_template contains variables, use input to specify
                                                            the values for those variables. Defaults to None, assuming no
                                                            variables.
                model (Optional[List[LLMConfig]]): List of models to choose from.
                default (Optional[Union[LLMConfig, int, str]]): Default LLM.
                max_model_depth (Optional[int]): If your top recommended model is down, specify up to which depth
                                                    of routing you're willing to go.
                latency_tracking (Optional[bool]): Latency tracking flag.
                hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
                tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                                    for the router to determine the best LLM for a given query.
                preference_id (Optional[str]): The ID of the router preference that was configured via the Dashboard.
                                                Defaults to None.
                metric (Metric, optional): Metric used by NotDiamond router to choose the best LLM.
                                                Defaults to Metric("accuracy").
                previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
                timeout (int): The number of seconds to wait before terminating the API call to Not Diamond backend.
                max_retries (int): The number of retries to attempt before giving up.
                **kwargs: Any other arguments that are supported by Langchain's invoke method, will be passed through.

            Returns:
                tuple[str, Optional[LLMConfig]]: returns the session_id and the chosen LLM
            """
            if input is None:
                input = {}

            if model is not None:
                llm_configs = self._parse_llm_configs_data(model)
                self.llm_configs = llm_configs

            self.validate_params(
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
            )

            best_llm, session_id = model_select(
                messages=messages,
                llm_configs=self.llm_configs,
                metric=metric,
                notdiamond_api_key=self.api_key,
                max_model_depth=self.max_model_depth,
                hash_content=self.hash_content,
                tradeoff=self.tradeoff,
                preference_id=self.preference_id,
                tools=self.tools,
                previous_session=previous_session,
                timeout=timeout or self.timeout,
                max_retries=max_retries or self.max_retries,
                nd_api_url=self.nd_api_url,
                _user_agent=self.user_agent,
            )

            if not best_llm:
                LOGGER.warning(
                    f"ND API error. Falling back to default provider={self.default_llm.provider}/{self.default_llm.model}."
                )
                best_llm = self.default_llm
            self.call_callbacks("on_model_select", best_llm, best_llm.model)

            return session_id, best_llm

        @staticmethod
        def _parse_llm_configs_data(
            llm_configs: list,
        ) -> List[LLMConfig]:
            providers = []
            for llm_config in llm_configs:
                if isinstance(llm_config, LLMConfig):
                    providers.append(llm_config)
                    continue
                parsed_provider = LLMConfig.from_string(llm_config)
                providers.append(parsed_provider)
            return providers

        def validate_params(
            self,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
        ):
            self.default = default

            if max_model_depth is not None:
                self.max_model_depth = max_model_depth

            if self.llm_configs is None or len(self.llm_configs) == 0:
                raise MissingLLMConfigs(
                    "No LLM config speficied. Specify at least one."
                )

            if self.max_model_depth is None:
                self.max_model_depth = len(self.llm_configs)

            if self.max_model_depth == 0:
                raise ValueError("max_model_depth has to be bigger than 0.")

            if self.max_model_depth > len(self.llm_configs):
                LOGGER.warning(
                    "WARNING: max_model_depth cannot be bigger than the number of LLMs."
                )
                self.max_model_depth = len(self.llm_configs)

            if tradeoff is not None:
                if tradeoff not in ["cost", "latency"]:
                    raise ValueError(
                        "Invalid tradeoff. Accepted values: cost, latency."
                    )
            self.tradeoff = tradeoff

            if preference_id is not None:
                self.preference_id = preference_id

            if latency_tracking is not None:
                self.latency_tracking = latency_tracking

            if hash_content is not None:
                self.hash_content = hash_content

        def bind_tools(
            self, tools: Sequence[Union[Dict[str, Any], Callable]]
        ) -> "NotDiamond":
            """
            Bind tools to the LLM object. The tools will be passed to the LLM object when invoking it.
            Results in the tools being available in the LLM object.
            You can access the tool_calls in the result via `result.tool_calls`.
            """

            for provider in self.llm_configs:
                if provider.model not in settings.PROVIDERS[
                    provider.provider
                ].get("support_tools", []):
                    raise ApiError(
                        f"{provider.provider}/{provider.model} does not support function calling."
                    )
            self.tools = tools

            return self

        def call_callbacks(self, function_name: str, *args, **kwargs) -> None:
            """
            Call all callbacks with a specific function name.
            """

            if self.callbacks is None:
                return

            for callback in self.callbacks:
                if hasattr(callback, function_name):
                    getattr(callback, function_name)(*args, **kwargs)

        def create(*args, **kwargs):
            format_str = f"`{inspect.stack()[0].function}`"
            raise CreateUnavailableError(
                _invoke_error_msg_tmpl.format(fn_name=format_str)
            )

        async def acreate(*args, **kwargs):
            format_str = f"`{inspect.stack()[0].function}`"
            raise CreateUnavailableError(
                _invoke_error_msg_tmpl.format(fn_name=format_str)
            )

        def invoke(*args, **kwargs):
            format_str = f"`{inspect.stack()[0].function}`"
            raise CreateUnavailableError(
                _invoke_error_msg_tmpl.format(fn_name=format_str)
            )

        async def ainvoke(*args, **kwargs):
            format_str = f"`{inspect.stack()[0].function}`"
            raise CreateUnavailableError(
                _invoke_error_msg_tmpl.format(fn_name=format_str)
            )

        def stream(*args, **kwargs):
            raise CreateUnavailableError(
                _invoke_error_msg_tmpl.format(
                    fn_name=inspect.stack()[0].function
                )
            )

        async def astream(*args, **kwargs):
            raise CreateUnavailableError(
                _invoke_error_msg_tmpl.format(
                    fn_name=inspect.stack()[0].function
                )
            )

        @property
        def default_llm(self) -> LLMConfig:
            """
            Return the default LLM that's set on the NotDiamond client class.
            """
            if isinstance(self.default, int):
                if self.default < len(self.llm_configs):
                    return self.llm_configs[self.default]

            if isinstance(self.default, str):
                try:
                    default = LLMConfig.from_string(self.default)
                    if default in self.llm_configs:
                        return default
                except Exception as e:
                    LOGGER.debug(f"Error setting default llm: {e}")

            if isinstance(self.default, LLMConfig):
                return self.default

            default = self.llm_configs[0]
            if self.default is None:
                LOGGER.info(
                    _no_default_llm_config_warning.format(
                        provider=default.provider, model=default.model
                    )
                )
            else:
                LOGGER.info(
                    _default_llm_config_invalid_warning.format(
                        provider=default.provider, model=default.model
                    )
                )
            return default

    # Do not import from langchain_core directly, as it is now an optional SDK dependency
    try:
        LLM = _module_check("langchain_core.language_models.llms", "LLM")
        BaseMessageChunk = _module_check(
            "langchain_core.messages", "BaseMessageChunk"
        )
        JsonOutputParser = _module_check(
            "langchain_core.output_parsers", "JsonOutputParser"
        )
        ChatPromptTemplate = _module_check(
            "langchain_core.prompts", "ChatPromptTemplate"
        )
    except (ModuleNotFoundError, ImportError) as ierr:
        msg = _invoke_error_msg_tmpl.format(fn_name="NotDiamond creation")
        if import_target == _NDClientTarget.INVOKER:
            msg += " Create was requested, however - raising..."
            raise ImportError(msg) from ierr
        else:
            LOGGER.debug(msg)
            return _NDRouterClient

    class _NDInvokerClient(_NDRouterClient, LLM):
        """
        Implementation of NotDiamond class, the main class responsible for creating and invoking LLM prompts.
        The class inherits from Langchain's LLM class. Starting reference is from here:
        https://python.langchain.com/docs/modules/model_io/llms/custom_llm

        It's mandatory to have an API key set. If the api_key is not explicitly specified,
        it will check for NOTDIAMOND_API_KEY in the .env file.

        Raises:
            MissingLLMProviders: you must specify at least one LLM provider for the router to work
            ApiError: error raised when the NotDiamond API call fails.
                        Ensure to set a default LLM provider to not break the code.
        """

        api_key: str
        llm_configs: Optional[List[Union[LLMConfig, str]]]
        default: Union[LLMConfig, int, str]
        max_model_depth: Optional[int]
        latency_tracking: bool
        hash_content: bool
        tradeoff: Optional[str]
        preference_id: Optional[str]
        tools: Optional[Sequence[Union[Dict[str, Any], Callable]]]
        callbacks: Optional[List]
        nd_api_url: Optional[str]
        user_agent: Union[str, None]

        def __init__(
            self,
            llm_configs: Optional[List[Union[LLMConfig, str]]] = None,
            api_key: Optional[str] = None,
            default: Union[LLMConfig, int, str] = 0,
            max_model_depth: Optional[int] = None,
            latency_tracking: bool = True,
            hash_content: bool = False,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            tools: Optional[Sequence[Union[Dict[str, Any], Callable]]] = None,
            callbacks: Optional[List] = None,
            nd_api_url: Optional[str] = settings.NOTDIAMOND_API_URL,
            user_agent: Union[str, None] = None,
            timeout: Optional[Union[float, int]] = 60.0,
            max_retries: Optional[int] = 3,
            **kwargs,
        ) -> None:
            super().__init__(
                api_key=api_key,
                llm_configs=llm_configs,
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
                tools=tools,
                callbacks=callbacks,
                nd_api_url=nd_api_url,
                user_agent=user_agent,
                timeout=timeout,
                max_retries=max_retries,
                **kwargs,
            )
            if user_agent is None:
                user_agent = settings.DEFAULT_USER_AGENT

            if tradeoff is not None:
                warnings.warn(
                    "The tradeoff constructor parameter is deprecated and will be removed in a "
                    "future version. Please specify the tradeoff when using model_select or invocation methods.",
                    DeprecationWarning,
                    stacklevel=2,
                )

            self.user_agent = user_agent
            assert (
                self.api_key is not None
            ), "API key is not set. Please set a Not Diamond API key."

        def __repr__(self) -> str:
            class_name = self.__class__.__name__
            address = hex(id(self))  # Gets the memory address of the object
            return f"<{class_name} object at {address}>"

        @property
        def _llm_type(self) -> str:
            return "NotDiamond LLM"

        @staticmethod
        def _inject_model_instruction(messages, parser):
            format_instructions = parser.get_format_instructions()
            format_instructions = format_instructions.replace(
                "{", "{{"
            ).replace("}", "}}")
            messages[0]["content"] = (
                format_instructions + "\n" + messages[0]["content"]
            )
            return messages

        def _call(
            self,
            prompt: str,
            stop: Optional[List[str]] = None,
            run_manager: Optional[Any] = None,
            **kwargs: Any,
        ) -> str:
            if stop is not None:
                raise ValueError("stop kwargs are not permitted.")
            return "This function is deprecated for the latest LangChain version, use invoke instead"

        def create(
            self,
            messages: List[Dict[str, str]],
            model: Optional[List[LLMConfig]] = None,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            metric: Metric = Metric("accuracy"),
            previous_session: Optional[str] = None,
            response_model: Optional[Type[BaseModel]] = None,
            timeout: Optional[Union[float, int]] = None,
            max_retries: Optional[int] = None,
            **kwargs,
        ) -> tuple[str, str, LLMConfig]:
            """
            Function call to invoke the LLM, with the same interface
            as the OpenAI Python library.

            Parameters:
                messages (Optional[List[Dict[str, str]], optional): Can be used instead of prompt_template to pass
                    the messages OpenAI style.
                model (Optional[List[LLMConfig]]): List of models to choose from.
                default (Optional[Union[LLMConfig, int, str]]): Default LLM.
                max_model_depth (Optional[int]): If your top recommended model is down, specify up to which depth
                                                    of routing you're willing to go.
                latency_tracking (Optional[bool]): Latency tracking flag.
                hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
                tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                                    for the router to determine the best LLM for a given query.
                preference_id (Optional[str]): The ID of the router preference that was configured via the Dashboard.
                                                Defaults to None.
                metric (Metric, optional): Metric used by NotDiamond router to choose the best LLM.
                                                Defaults to Metric("accuracy").
                previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
                response_model (Optional[Type[BaseModel]], optional): If present, will use JsonOutputParser to parse the
                                                                response into the given model. In which case result will a
                                                                dict.
                timeout (int): The number of seconds to wait before terminating the API call to Not Diamond backend.
                max_retries (int): The number of retries to attempt before giving up.
                nd_api_url (Optional[str]): The URL of the NotDiamond API. Defaults to settings.NOTDIAMOND_API_URL.
                **kwargs: Any other arguments that are supported by Langchain's invoke method, will be passed through.

            Raises:
                ApiError: when the NotDiamond API fails

            Returns:
                tuple[Union[AIMessage, BaseModel], str, LLMConfig]:
                    result: response type defined by Langchain, contains the response from the LLM.
                    or object of the response_model
                    str: session_id returned by the NotDiamond API
                    LLMConfig: the best LLM selected by the router
            """

            return self.invoke(
                messages=messages,
                model=model,
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
                metric=metric,
                previous_session=previous_session,
                response_model=response_model,
                timeout=timeout,
                max_retries=max_retries,
                **kwargs,
            )

        async def acreate(
            self,
            messages: List[Dict[str, str]],
            model: Optional[List[LLMConfig]] = None,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            metric: Metric = Metric("accuracy"),
            previous_session: Optional[str] = None,
            response_model: Optional[Type[BaseModel]] = None,
            timeout: Optional[Union[float, int]] = None,
            max_retries: Optional[int] = None,
            **kwargs,
        ) -> tuple[str, str, LLMConfig]:
            """
            Async function call to invoke the LLM, with the same interface
            as the OpenAI Python library.

            Parameters:
                messages (Optional[List[Dict[str, str]], optional): Can be used instead of prompt_template to pass
                    the messages OpenAI style.
                model (Optional[List[LLMConfig]]): List of models to choose from.
                default (Optional[Union[LLMConfig, int, str]]): Default LLM.
                max_model_depth (Optional[int]): If your top recommended model is down, specify up to which depth
                                                    of routing you're willing to go.
                latency_tracking (Optional[bool]): Latency tracking flag.
                hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
                tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                                    for the router to determine the best LLM for a given query.
                preference_id (Optional[str]): The ID of the router preference that was configured via the Dashboard.
                                                Defaults to None.
                metric (Metric, optional): Metric used by NotDiamond router to choose the best LLM.
                                                Defaults to Metric("accuracy").
                previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
                response_model (Optional[Type[BaseModel]], optional): If present, will use JsonOutputParser to parse the
                                                                response into the given model. In which case result will a
                                                                dict.
                timeout (int): The number of seconds to wait before terminating the API call to Not Diamond backend.
                max_retries (int): The number of retries to attempt before giving up.
                **kwargs: Any other arguments that are supported by Langchain's invoke method, will be passed through.

            Raises:
                ApiError: when the NotDiamond API fails

            Returns:
                tuple[Union[AIMessage, BaseModel], str, LLMConfig]:
                    result: response type defined by Langchain, contains the response from the LLM.
                    or object of the response_model
                    str: session_id returned by the NotDiamond API
                    LLMConfig: the best LLM selected by the router
            """

            result = await self.ainvoke(
                messages=messages,
                model=model,
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
                metric=metric,
                previous_session=previous_session,
                response_model=response_model,
                timeout=timeout,
                max_retries=max_retries,
                **kwargs,
            )
            return result

        def invoke(
            self,
            messages: List[Dict[str, str]],
            model: Optional[List[LLMConfig]] = None,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            metric: Metric = Metric("accuracy"),
            previous_session: Optional[str] = None,
            response_model: Optional[Type[BaseModel]] = None,
            timeout: Optional[Union[float, int]] = None,
            max_retries: Optional[int] = None,
            input: Optional[Dict[str, Any]] = None,
            **kwargs,
        ) -> tuple[str, str, LLMConfig]:
            """
            Function to invoke the LLM. Behind the scenes what happens:
            1. API call to NotDiamond backend to get the most suitable LLM for the given prompt
            2. Invoke the returned LLM client side
            3. Return the response

            Parameters:
                prompt_template (Optional(Union[ NDPromptTemplate, NDChatPromptTemplate, str, ])):
                    the prompt template defined by the user. It also supports Langchain prompt template types.
                messages (Optional[List[Dict[str, str]], optional): Can be used instead of prompt_template to pass
                    the messages OpenAI style.
                model (Optional[List[LLMConfig]]): List of models to choose from.
                default (Optional[Union[LLMConfig, int, str]]): Default LLM.
                max_model_depth (Optional[int]): If your top recommended model is down, specify up to which depth
                                                    of routing you're willing to go.
                latency_tracking (Optional[bool]): Latency tracking flag.
                hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
                tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                                    for the router to determine the best LLM for a given query.
                preference_id (Optional[str]): The ID of the router preference that was configured via the Dashboard.
                                                Defaults to None.
                metric (Metric, optional): Metric used by NotDiamond router to choose the best LLM.
                                                Defaults to Metric("accuracy").
                previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
                response_model (Optional[Type[BaseModel]], optional): If present, will use JsonOutputParser to parse the
                                                                response into the given model. In which case result will a
                                                                dict.
                timeout (int): The number of seconds to wait before terminating the API call to Not Diamond backend.
                max_retries (int): The number of retries to attempt before giving up.
                input (Optional[Dict[str, Any]], optional): If the prompt_template contains variables, use input to specify
                                                            the values for those variables. Defaults to None, assuming no
                                                            variables.
                **kwargs: Any other arguments that are supported by Langchain's invoke method, will be passed through.

            Raises:
                ApiError: when the NotDiamond API fails

            Returns:
                tuple[Union[AIMessage, BaseModel], str, LLMConfig]:
                    result: response type defined by Langchain, contains the response from the LLM.
                    or object of the response_model
                    str: session_id returned by the NotDiamond API
                    LLMConfig: the best LLM selected by the router
            """

            if model is not None:
                llm_configs = self._parse_llm_configs_data(model)
                self.llm_configs = llm_configs

            self.validate_params(
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
            )

            # If response_model is present, we will parse the response into the given model
            # doing this here so that if validation errors occur, we can raise them before making the API call
            response_model_parser = None
            if response_model is not None:
                self.verify_against_response_model()
                response_model_parser = JsonOutputParser(
                    pydantic_object=response_model
                )

            if input is None:
                input = {}

            best_llm, session_id = model_select(
                messages=messages,
                llm_configs=self.llm_configs,
                metric=metric,
                notdiamond_api_key=self.api_key,
                max_model_depth=self.max_model_depth,
                hash_content=self.hash_content,
                tradeoff=self.tradeoff,
                preference_id=self.preference_id,
                tools=self.tools,
                previous_session=previous_session,
                timeout=timeout or self.timeout,
                max_retries=max_retries or self.max_retries,
                nd_api_url=self.nd_api_url,
            )

            is_default = False
            if not best_llm:
                LOGGER.warning(
                    f"ND API error. Falling back to default provider={self.default_llm.provider}/{self.default_llm.model}."
                )
                best_llm = self.default_llm
                is_default = True

            if best_llm.system_prompt is not None:
                messages = inject_system_prompt(
                    messages, best_llm.system_prompt
                )

            messages = o1_system_prompt_translate(messages, best_llm)

            self.call_callbacks("on_model_select", best_llm, best_llm.model)

            llm = self._llm_from_config(best_llm, callbacks=self.callbacks)

            if self.tools and not is_o1_model(best_llm):
                llm = llm.bind_tools(self.tools)

            if response_model is not None:
                messages = _NDInvokerClient._inject_model_instruction(
                    messages, response_model_parser
                )
            chain_messages = [
                (msg["role"], _curly_escape(msg["content"]))
                for msg in messages
            ]
            prompt_template = ChatPromptTemplate.from_messages(chain_messages)
            chain = prompt_template | llm
            accepted_errors = _get_accepted_invoke_errors(best_llm.provider)

            try:
                if self.latency_tracking:
                    result = self._invoke_with_latency_tracking(
                        session_id=session_id,
                        chain=chain,
                        llm_config=best_llm,
                        is_default=is_default,
                        input=input,
                        **kwargs,
                    )
                else:
                    result = chain.invoke(input, **kwargs)
            except accepted_errors as e:
                if best_llm.provider == "google":
                    LOGGER.warning(
                        f"Submitted chat messages are violating Google requirements with error {e}. "
                        "If you see this message, `notdiamond` has returned a Google model as the best option, "
                        "but the LLM call will fail. If possible, `notdiamond` will fall back to a non-Google model."
                    )

                    non_google_llm = next(
                        (
                            llm_config
                            for llm_config in self.llm_configs
                            if llm_config.provider != "google"
                        ),
                        None,
                    )

                    if non_google_llm is not None:
                        best_llm = non_google_llm
                        llm = self._llm_from_config(
                            best_llm, callbacks=self.callbacks
                        )
                        if response_model is not None:
                            messages = (
                                _NDInvokerClient._inject_model_instruction(
                                    messages, response_model_parser
                                )
                            )
                        chain_messages = [
                            (msg["role"], _curly_escape(msg["content"]))
                            for msg in messages
                        ]
                        prompt_template = ChatPromptTemplate.from_messages(
                            chain_messages
                        )
                        chain = prompt_template | llm

                        if self.latency_tracking:
                            result = self._invoke_with_latency_tracking(
                                session_id=session_id,
                                chain=chain,
                                llm_config=best_llm,
                                is_default=is_default,
                                input=input,
                                **kwargs,
                            )
                        else:
                            result = chain.invoke(input, **kwargs)
                    else:
                        raise e
                else:
                    raise e

            if response_model is not None:
                parsed_dict = response_model_parser.parse(result.content)
                result = response_model.parse_obj(parsed_dict)

            return result, session_id, best_llm

        async def ainvoke(
            self,
            messages: List[Dict[str, str]],
            model: Optional[List[LLMConfig]] = None,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            metric: Metric = Metric("accuracy"),
            previous_session: Optional[str] = None,
            response_model: Optional[Type[BaseModel]] = None,
            timeout: Optional[Union[float, int]] = None,
            max_retries: Optional[int] = None,
            input: Optional[Dict[str, Any]] = None,
            **kwargs,
        ) -> tuple[str, str, LLMConfig]:
            """
            Function to invoke the LLM. Behind the scenes what happens:
            1. API call to NotDiamond backend to get the most suitable LLM for the given prompt
            2. Invoke the returned LLM client side
            3. Return the response

            Parameters:
                messages (List[Dict[str, str]]): List of messages, OpenAI style
                model (Optional[List[LLMConfig]]): List of models to choose from.
                default (Optional[Union[LLMConfig, int, str]]): Default LLM.
                max_model_depth (Optional[int]): If your top recommended model is down, specify up to which depth
                                                    of routing you're willing to go.
                latency_tracking (Optional[bool]): Latency tracking flag.
                hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
                tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                                    for the router to determine the best LLM for a given query.
                preference_id (Optional[str]): The ID of the router preference that was configured via the Dashboard.
                                                Defaults to None.
                metric (Metric, optional): Metric used by NotDiamond router to choose the best LLM.
                                                Defaults to Metric("accuracy").
                previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
                response_model (Optional[Type[BaseModel]], optional): If present, will use JsonOutputParser to parse the
                                                                response into the given model. In which case result will a dict.
                timeout (int): The number of seconds to wait before terminating the API call to Not Diamond backend.
                max_retries (int): The number of retries to attempt before giving up.
                input (Optional[Dict[str, Any]], optional): If the prompt_template contains variables, use input to specify
                                                            the values for those variables. Defaults to None, assuming no
                                                            variables.
                **kwargs: Any other arguments that are supported by Langchain's invoke method, will be passed through.

            Raises:
                ApiError: when the NotDiamond API fails

            Returns:
                tuple[Union[AIMessage, BaseModel], str, LLMConfig]:
                    result: response type defined by Langchain, contains the response from the LLM.
                    or object of the response_model
                    str: session_id returned by the NotDiamond API
                    LLMConfig: the best LLM selected by the router
            """

            if model is not None:
                llm_configs = self._parse_llm_configs_data(model)
                self.llm_configs = llm_configs

            self.validate_params(
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
            )

            response_model_parser = None
            if response_model is not None:
                self.verify_against_response_model()
                response_model_parser = JsonOutputParser(
                    pydantic_object=response_model
                )

            if input is None:
                input = {}

            best_llm, session_id = await amodel_select(
                messages=messages,
                llm_configs=self.llm_configs,
                metric=metric,
                notdiamond_api_key=self.api_key,
                max_model_depth=self.max_model_depth,
                hash_content=self.hash_content,
                tradeoff=self.tradeoff,
                preference_id=self.preference_id,
                tools=self.tools,
                previous_session=previous_session,
                timeout=timeout or self.timeout,
                max_retries=max_retries or self.max_retries,
                nd_api_url=self.nd_api_url,
            )

            is_default = False
            if not best_llm:
                LOGGER.warning(
                    f"ND API error. Falling back to default provider={self.default_llm.provider}/{self.default_llm.model}."
                )
                best_llm = self.default_llm
                is_default = True

            if best_llm.system_prompt is not None:
                messages = inject_system_prompt(
                    messages, best_llm.system_prompt
                )

            messages = o1_system_prompt_translate(messages, best_llm)

            self.call_callbacks("on_model_select", best_llm, best_llm.model)

            llm = self._llm_from_config(best_llm, callbacks=self.callbacks)

            if self.tools and not is_o1_model(best_llm):
                llm = llm.bind_tools(self.tools)

            if response_model is not None:
                messages = _NDInvokerClient._inject_model_instruction(
                    messages, response_model_parser
                )
            chain_messages = [
                (msg["role"], _curly_escape(msg["content"]))
                for msg in messages
            ]
            prompt_template = ChatPromptTemplate.from_messages(chain_messages)
            chain = prompt_template | llm
            accepted_errors = _get_accepted_invoke_errors(best_llm.provider)

            try:
                if self.latency_tracking:
                    result = await self._async_invoke_with_latency_tracking(
                        session_id=session_id,
                        chain=chain,
                        llm_config=best_llm,
                        is_default=is_default,
                        input=input,
                        **kwargs,
                    )
                else:
                    result = await chain.ainvoke(input, **kwargs)
            except accepted_errors as e:
                if best_llm.provider == "google":
                    LOGGER.warning(
                        f"Submitted chat messages are violating Google requirements with error {e}. "
                        "If you see this message, `notdiamond` has returned a Google model as the best option, "
                        "but the LLM call will fail. If possible, `notdiamond` will fall back to a non-Google model."
                    )

                    non_google_llm = next(
                        (
                            llm_config
                            for llm_config in self.llm_configs
                            if llm_config.provider != "google"
                        ),
                        None,
                    )

                    if non_google_llm is not None:
                        best_llm = non_google_llm
                        llm = self._llm_from_config(
                            best_llm, callbacks=self.callbacks
                        )
                        if response_model is not None:
                            messages = (
                                _NDInvokerClient._inject_model_instruction(
                                    messages, response_model_parser
                                )
                            )
                        chain_messages = [
                            (msg["role"], _curly_escape(msg["content"]))
                            for msg in messages
                        ]
                        prompt_template = ChatPromptTemplate.from_messages(
                            chain_messages
                        )
                        chain = prompt_template | llm

                        if self.latency_tracking:
                            result = (
                                await self._async_invoke_with_latency_tracking(
                                    session_id=session_id,
                                    chain=chain,
                                    llm_config=best_llm,
                                    is_default=is_default,
                                    input=input,
                                    **kwargs,
                                )
                            )
                        else:
                            result = await chain.ainvoke(input, **kwargs)
                    else:
                        raise e
                else:
                    raise e

            if response_model is not None:
                parsed_dict = response_model_parser.parse(result.content)
                result = response_model.parse_obj(parsed_dict)

            return result, session_id, best_llm

        def stream(
            self,
            messages: List[Dict[str, str]],
            model: Optional[List[LLMConfig]] = None,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            metric: Metric = Metric("accuracy"),
            previous_session: Optional[str] = None,
            response_model: Optional[Type[BaseModel]] = None,
            timeout: Optional[Union[float, int]] = None,
            max_retries: Optional[int] = None,
            **kwargs,
        ) -> Iterator[Union[BaseMessageChunk, BaseModel]]:
            """
            This function calls the NotDiamond backend to fetch the most suitable model for the given prompt,
            and calls the LLM client side to stream the response.

            Parameters:
                messages (Optional[List[Dict[str, str]], optional): List of messages, OpenAI style
                model (Optional[List[LLMConfig]]): List of models to choose from.
                default (Optional[Union[LLMConfig, int, str]]): Default LLM.
                max_model_depth (Optional[int]): If your top recommended model is down, specify up to which depth
                                                    of routing you're willing to go.
                latency_tracking (Optional[bool]): Latency tracking flag.
                hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
                tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                                    for the router to determine the best LLM for a given query.
                preference_id (Optional[str]): The ID of the router preference that was configured via the Dashboard.
                                                Defaults to None.
                metric (Metric, optional): Metric used by NotDiamond router to choose the best LLM.
                                                Defaults to Metric("accuracy").
                previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
                response_model (Optional[Type[BaseModel]], optional): If present, will use JsonOutputParser to parse the
                                                                response into the given model. In which case result will a
                                                                dict.
                timeout (int): The number of seconds to wait before terminating the API call to Not Diamond backend.
                max_retries (int): The number of retries to attempt before giving up.
                **kwargs: Any other arguments that are supported by Langchain's invoke method, will be passed through.

            Raises:
                ApiError: when the NotDiamond API fails

            Yields:
                Iterator[Union[BaseMessageChunk, BaseModel]]: returns the response in chunks.
                    If response_model is present, it will return the partial model object
            """

            if model is not None:
                llm_configs = self._parse_llm_configs_data(model)
                self.llm_configs = llm_configs

            self.validate_params(
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
            )

            response_model_parser = None
            if response_model is not None:
                self.verify_against_response_model()
                response_model_parser = JsonOutputParser(
                    pydantic_object=response_model
                )

            best_llm, session_id = model_select(
                messages=messages,
                llm_configs=self.llm_configs,
                metric=metric,
                notdiamond_api_key=self.api_key,
                max_model_depth=self.max_model_depth,
                hash_content=self.hash_content,
                tradeoff=self.tradeoff,
                preference_id=self.preference_id,
                tools=self.tools,
                previous_session=previous_session,
                timeout=timeout or self.timeout,
                max_retries=max_retries or self.max_retries,
                nd_api_url=self.nd_api_url,
            )

            if not best_llm:
                LOGGER.warning(
                    f"ND API error. Falling back to default provider={self.default_llm.provider}/{self.default_llm.model}."
                )
                best_llm = self.default_llm

            if best_llm.system_prompt is not None:
                messages = inject_system_prompt(
                    messages, best_llm.system_prompt
                )

            if response_model is not None:
                messages = _NDInvokerClient._inject_model_instruction(
                    messages, response_model_parser
                )

            self.call_callbacks("on_model_select", best_llm, best_llm.model)

            llm = self._llm_from_config(best_llm, callbacks=self.callbacks)
            if self.tools:
                llm = llm.bind_tools(self.tools)

            if response_model is not None:
                chain = llm | response_model_parser
            else:
                chain = llm

            for chunk in chain.stream(messages, **kwargs):
                if response_model is None:
                    yield chunk
                else:
                    partial_model = create_partial_model(response_model)
                    yield partial_model(**chunk)

        async def astream(
            self,
            messages: List[Dict[str, str]],
            model: Optional[List[LLMConfig]] = None,
            default: Optional[Union[LLMConfig, int, str]] = None,
            max_model_depth: Optional[int] = None,
            latency_tracking: Optional[bool] = None,
            hash_content: Optional[bool] = None,
            tradeoff: Optional[str] = None,
            preference_id: Optional[str] = None,
            metric: Metric = Metric("accuracy"),
            previous_session: Optional[str] = None,
            response_model: Optional[Type[BaseModel]] = None,
            timeout: Optional[Union[float, int]] = None,
            max_retries: Optional[int] = None,
            **kwargs,
        ) -> AsyncIterator[Union[BaseMessageChunk, BaseModel]]:
            """
            This function calls the NotDiamond backend to fetch the most suitable model for the given prompt,
            and calls the LLM client side to stream the response. The function is async, so it's suitable for async codebases.

            Parameters:
                messages (Optional[List[Dict[str, str]], optional): List of messages, OpenAI style
                model (Optional[List[LLMConfig]]): List of models to choose from.
                default (Optional[Union[LLMConfig, int, str]]): Default LLM.
                max_model_depth (Optional[int]): If your top recommended model is down, specify up to which depth
                                                    of routing you're willing to go.
                latency_tracking (Optional[bool]): Latency tracking flag.
                hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
                tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                                    for the router to determine the best LLM for a given query.
                preference_id (Optional[str]): The ID of the router preference that was configured via the Dashboard.
                                                Defaults to None.
                metric (Metric, optional): Metric used by NotDiamond router to choose the best LLM.
                                                Defaults to Metric("accuracy").
                previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
                response_model (Optional[Type[BaseModel]], optional): If present, will use JsonOutputParser to parse the
                                                                response into the given model. In which case result will a dict.
                timeout (int): The number of seconds to wait before terminating the API call to Not Diamond backend.
                max_retries (int): The number of retries to attempt before giving up.
                **kwargs: Any other arguments that are supported by Langchain's invoke method, will be passed through.

            Raises:
                ApiError: when the NotDiamond API fails

            Yields:
                AsyncIterator[Union[BaseMessageChunk, BaseModel]]: returns the response in chunks.
                    If response_model is present, it will return the partial model object
            """

            if model is not None:
                llm_configs = self._parse_llm_configs_data(model)
                self.llm_configs = llm_configs

            self.validate_params(
                default=default,
                max_model_depth=max_model_depth,
                latency_tracking=latency_tracking,
                hash_content=hash_content,
                tradeoff=tradeoff,
                preference_id=preference_id,
            )

            response_model_parser = None
            if response_model is not None:
                self.verify_against_response_model()
                response_model_parser = JsonOutputParser(
                    pydantic_object=response_model
                )

            best_llm, session_id = await amodel_select(
                messages=messages,
                llm_configs=self.llm_configs,
                metric=metric,
                notdiamond_api_key=self.api_key,
                max_model_depth=self.max_model_depth,
                hash_content=self.hash_content,
                tradeoff=self.tradeoff,
                preference_id=self.preference_id,
                tools=self.tools,
                previous_session=previous_session,
                timeout=timeout or self.timeout,
                max_retries=max_retries or self.max_retries,
                nd_api_url=self.nd_api_url,
            )

            if not best_llm:
                LOGGER.warning(
                    f"ND API error. Falling back to default provider={self.default_llm.provider}/{self.default_llm.model}."
                )
                best_llm = self.default_llm

            if best_llm.system_prompt is not None:
                messages = inject_system_prompt(
                    messages, best_llm.system_prompt
                )
            if response_model is not None:
                messages = _NDInvokerClient._inject_model_instruction(
                    messages, response_model_parser
                )

            self.call_callbacks("on_model_select", best_llm, best_llm.model)

            llm = self._llm_from_config(best_llm, callbacks=self.callbacks)
            if self.tools:
                llm = llm.bind_tools(self.tools)

            if response_model is not None:
                chain = llm | response_model_parser
            else:
                chain = llm

            async for chunk in chain.astream(messages, **kwargs):
                if response_model is None:
                    yield chunk
                else:
                    partial_model = create_partial_model(response_model)
                    yield partial_model(**chunk)

        async def _async_invoke_with_latency_tracking(
            self,
            session_id: str,
            chain: Any,
            llm_config: LLMConfig,
            input: Optional[Dict[str, Any]] = {},
            is_default: bool = True,
            **kwargs,
        ):
            if session_id in ("NO-SESSION-ID", "") and not is_default:
                error_message = (
                    "ND session_id is not valid for latency tracking."
                    + "Please check the API response."
                )
                self.call_callbacks("on_api_error", error_message)
                raise ApiError(error_message)

            start_time = time.time()

            result = await chain.ainvoke(input, **kwargs)

            end_time = time.time()

            tokens_completed = token_counter(
                model=llm_config.model,
                messages=[{"role": "assistant", "content": result.content}],
            )
            tokens_per_second = tokens_completed / (end_time - start_time)

            report_latency(
                session_id=session_id,
                llm_config=llm_config,
                tokens_per_second=tokens_per_second,
                notdiamond_api_key=self.api_key,
                nd_api_url=self.nd_api_url,
                _user_agent=self.user_agent,
            )
            self.call_callbacks(
                "on_latency_tracking",
                session_id,
                llm_config,
                tokens_per_second,
            )

            return result

        def _invoke_with_latency_tracking(
            self,
            session_id: str,
            chain: Any,
            llm_config: LLMConfig,
            input: Optional[Dict[str, Any]] = {},
            is_default: bool = True,
            **kwargs,
        ):
            LOGGER.debug(f"Latency tracking enabled, session_id={session_id}")
            if session_id in ("NO-SESSION-ID", "") and not is_default:
                error_message = (
                    "ND session_id is not valid for latency tracking."
                    + "Please check the API response."
                )
                self.call_callbacks("on_api_error", error_message)
                raise ApiError(error_message)

            start_time = time.time()
            result = chain.invoke(input, **kwargs)
            end_time = time.time()

            tokens_completed = token_counter(
                model=llm_config.model,
                messages=[{"role": "assistant", "content": result.content}],
            )
            tokens_per_second = tokens_completed / (end_time - start_time)

            report_latency(
                session_id=session_id,
                llm_config=llm_config,
                tokens_per_second=tokens_per_second,
                notdiamond_api_key=self.api_key,
                nd_api_url=self.nd_api_url,
                _user_agent=self.user_agent,
            )
            self.call_callbacks(
                "on_latency_tracking",
                session_id,
                llm_config,
                tokens_per_second,
            )

            return result

        @staticmethod
        def _llm_from_config(
            provider: LLMConfig,
            callbacks: Optional[List] = None,
        ) -> Any:
            default_kwargs = {"max_retries": 5, "timeout": 120}
            passed_kwargs = {**default_kwargs, **provider.kwargs}

            if provider.provider == "openai":
                ChatOpenAI = _module_check(
                    "langchain_openai.chat_models",
                    "ChatOpenAI",
                    provider.provider,
                )
                if is_o1_model(provider):
                    passed_kwargs["temperature"] = 1.0

                return ChatOpenAI(
                    openai_api_key=provider.api_key,
                    model_name=provider.model,
                    callbacks=callbacks,
                    **passed_kwargs,
                )
            if provider.provider == "anthropic":
                ChatAnthropic = _module_check(
                    "langchain_anthropic", "ChatAnthropic", provider.provider
                )
                return ChatAnthropic(
                    anthropic_api_key=provider.api_key,
                    model=provider.model,
                    callbacks=callbacks,
                    **passed_kwargs,
                )
            if provider.provider == "google":
                ChatGoogleGenerativeAI = _module_check(
                    "langchain_google_genai",
                    "ChatGoogleGenerativeAI",
                    provider.provider,
                )
                return ChatGoogleGenerativeAI(
                    google_api_key=provider.api_key,
                    model=provider.model,
                    convert_system_message_to_human=True,
                    callbacks=callbacks,
                    **passed_kwargs,
                )
            if provider.provider == "cohere":
                ChatCohere = _module_check(
                    "langchain_cohere.chat_models",
                    "ChatCohere",
                    provider.provider,
                )
                return ChatCohere(
                    cohere_api_key=provider.api_key,
                    model=provider.model,
                    callbacks=callbacks,
                    **passed_kwargs,
                )
            if provider.provider == "mistral":
                ChatMistralAI = _module_check(
                    "langchain_mistralai.chat_models",
                    "ChatMistralAI",
                    provider.provider,
                )
                return ChatMistralAI(
                    mistral_api_key=provider.api_key,
                    model=provider.model,
                    callbacks=callbacks,
                    **passed_kwargs,
                )
            if provider.provider == "togetherai":
                provider_settings = settings.PROVIDERS.get(
                    provider.provider, None
                )
                model_prefixes = provider_settings.get("model_prefix", None)
                model_prefix = model_prefixes.get(provider.model, None)
                del passed_kwargs["max_retries"]
                del passed_kwargs["timeout"]

                if model_prefix is not None:
                    model = f"{model_prefix}/{provider.model}"
                ChatTogether = _module_check(
                    "langchain_together", "ChatTogether", provider.provider
                )
                return ChatTogether(
                    together_api_key=provider.api_key,
                    model=model,
                    callbacks=callbacks,
                    **passed_kwargs,
                )
            if provider.provider == "perplexity":
                del passed_kwargs["max_retries"]
                passed_kwargs["request_timeout"] = passed_kwargs["timeout"]
                del passed_kwargs["timeout"]
                ChatPerplexity = _module_check(
                    "langchain_community.chat_models",
                    "ChatPerplexity",
                    provider.provider,
                )
                return ChatPerplexity(
                    pplx_api_key=provider.api_key,
                    model=provider.model,
                    callbacks=callbacks,
                    **passed_kwargs,
                )
            if provider.provider == "replicate":
                provider_settings = settings.PROVIDERS.get(
                    provider.provider, None
                )
                model_prefixes = provider_settings.get("model_prefix", None)
                model_prefix = model_prefixes.get(provider.model, None)
                passed_kwargs["request_timeout"] = passed_kwargs["timeout"]
                del passed_kwargs["timeout"]

                if model_prefix is not None:
                    model = f"replicate/{model_prefix}/{provider.model}"
                ChatLiteLLM = _module_check(
                    "langchain_community.chat_models",
                    "ChatLiteLLM",
                    provider.provider,
                )
                return ChatLiteLLM(
                    model=model,
                    callbacks=callbacks,
                    replicate_api_key=provider.api_key,
                    **passed_kwargs,
                )
            raise ValueError(f"Unsupported provider: {provider.provider}")

        def verify_against_response_model(self) -> bool:
            """
            Verify that the LLMs support response modeling.
            """

            for provider in self.llm_configs:
                if provider.model not in settings.PROVIDERS[
                    provider.provider
                ].get("support_response_model", []):
                    raise ApiError(
                        f"{provider.provider}/{provider.model} does not support response modeling."
                    )

            return True

    if import_target is _NDClientTarget.ROUTER:
        return _NDRouterClient
    return _NDInvokerClient


_NDClient = _ndllm_factory()


class NotDiamond(_NDClient):
    api_key: str
    """
    API key required for making calls to NotDiamond.
    You can get an API key via our dashboard: https://app.notdiamond.ai
    If an API key is not set, it will check for NOTDIAMOND_API_KEY in .env file.
    """

    llm_configs: Optional[List[Union[LLMConfig, str]]]
    """The list of LLMs that are available to route between."""

    default: Union[LLMConfig, int, str]
    """
    Set a default LLM, so in case anything goes wrong in the flow,
    as for example NotDiamond API call fails, your code won't break and you have
    a fallback model. There are various ways to configure a default model:

    - Integer, specifying the index of the default provider from the llm_configs list
    - String, similar how you can specify llm_configs, of structure 'provider_name/model_name'
    - LLMConfig, just directly specify the object of the provider

    By default, we will set your first LLM in the list as the default.
    """

    max_model_depth: Optional[int]
    """
    If your top recommended model is down, specify up to which depth of routing you're willing to go.
    If max_model_depth is not set, it defaults to the length of the llm_configs list.
    If max_model_depth is set to 0, the init will fail.
    If the value is larger than the llm_configs list length, we reset the value to len(llm_configs).
    """

    latency_tracking: bool
    """
    Tracking and sending latency of LLM call to NotDiamond server as feedback, so we can improve our router.
    By default this is turned on, set it to False to turn off.
    """

    hash_content: bool
    """
    Hashing the content before being sent to the NotDiamond API.
    By default this is False.
    """

    tradeoff: Optional[str]
    """
    [DEPRECATED] The tradeoff constructor parameter is deprecated and will be removed in a future version.
    Please specify the tradeoff when using model_select or invocation methods.

    Define tradeoff between "cost" and "latency" for the router to determine the best LLM for a given query.
    If None is specified, then the router will not consider either cost or latency.

    The supported values: "cost", "latency"

    Defaults to None.
    """

    preference_id: Optional[str]
    """The ID of the router preference that was configured via the Dashboard. Defaults to None."""

    tools: Optional[Sequence[Union[Dict[str, Any], Callable]]]
    """Bind tools to the LLM object. The tools will be passed to the LLM object when invoking it."""

    nd_api_url: Optional[str]
    """The URL of the NotDiamond API. Defaults to settings.NOTDIAMOND_API_URL."""

    user_agent: Union[str, None]

    max_retries: int
    """The maximum number of retries to make when calling the Not Diamond API."""

    timeout: float
    """The timeout for the Not Diamond API call."""

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        nd_api_url: Optional[str] = settings.NOTDIAMOND_API_URL,
        user_agent: Union[str, None] = settings.DEFAULT_USER_AGENT,
        *args,
        **kwargs,
    ):
        super().__init__(
            nd_api_url=nd_api_url, user_agent=user_agent, *args, **kwargs
        )
        self.nd_api_url = nd_api_url

        if kwargs.get("tradeoff") is not None:
            warnings.warn(
                "The tradeoff constructor parameter is deprecated and will be removed in a "
                "future version. Please specify the tradeoff when using model_select or invocation methods.",
                DeprecationWarning,
                stacklevel=2,
            )


def _get_accepted_invoke_errors(provider: str) -> Tuple:
    if provider == "google":
        ChatGoogleGenerativeAIError = _module_check(
            "langchain_google_genai.chat_models",
            "ChatGoogleGenerativeAIError",
            provider,
        )
        accepted_errors = (ChatGoogleGenerativeAIError, ValueError)
    else:
        accepted_errors = (ValueError,)
    return accepted_errors



from ragas.metrics import (  # noqa
    AspectCritic,
    BleuScore,
    ContextEntityRecall,
    ContextPrecision,
    ContextRecall,
    ExactMatch,
    FactualCorrectness,
    Faithfulness,
    FaithfulnesswithHHEM,
    InstanceRubrics,
)
from ragas.metrics import (  # noqa
    LLMContextPrecisionWithoutReference as ContextPrecisionWithoutReference,
)
from ragas.metrics import (  # noqa
    LLMContextPrecisionWithReference,
    LLMContextRecall,
    NoiseSensitivity,
    NonLLMContextPrecisionWithReference,
    NonLLMContextRecall,
    NonLLMStringSimilarity,
    ResponseRelevancy,
    RougeScore,
    RubricsScore,
    SemanticSimilarity,
    SimpleCriteriaScore,
    StringPresence,
)



import logging
from typing import Dict, List, Optional, Sequence, Tuple, Union

import optuna
import pandas as pd
from langchain_core.callbacks import Callbacks
from langchain_core.embeddings import Embeddings as LangchainEmbeddings
from langchain_core.language_models import BaseLanguageModel as LangchainLLM
from langchain_core.prompt_values import StringPromptValue
from ragas import EvaluationDataset, SingleTurnSample
from ragas._analytics import track_was_completed
from ragas.cost import TokenUsageParser
from ragas.embeddings.base import BaseRagasEmbeddings
from ragas.evaluation import evaluate as ragas_evaluate
from ragas.llms import BaseRagasLLM
from ragas.metrics.base import Metric
from ragas.run_config import RunConfig
from tqdm import tqdm

from notdiamond.llms.config import LLMConfig
from notdiamond.toolkit.rag.evaluation_dataset import (
    RAGEvaluationDataset,
    RAGSample,
)
from notdiamond.toolkit.rag.llms import get_llm
from notdiamond.toolkit.rag.workflow import BaseNDRagWorkflow

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

_DEFAULT_GENERATION_TEMPERATURE = 0.7


def get_eval_dataset(
    test_queries: pd.DataFrame,
    workflow: BaseNDRagWorkflow,
    generation_prompt: str = None,
    generator_llm: Union[LLMConfig, str] = None,
):
    """
    Create a dataset of RAGSample objects to evaluate the performance of a RAG workflow.

    Args:
        test_queries: A pandas DataFrame with schema implied by the method below.
        workflow: BaseNDRagWorkflow subclass created by the user.

    Schema for test_queries can be found below
    https://docs.ragas.io/en/stable/references/evaluation_schema/#ragas.dataset_schema.SingleTurnSample
    """
    samples = []
    for _, row in test_queries.iterrows():
        query = row["user_input"]
        reference = row["reference"]
        generation_prompt = generation_prompt or row.get("generation_prompt")
        if generation_prompt is None:
            raise ValueError(
                "Provided test queries DataFrame does not include"
                " 'generation_prompt' and 'generation_prompt' is not provided."
            )
        generator_llm = generator_llm or row.get("generator_llm")
        if generator_llm is None:
            raise ValueError(
                "Provided test queries DataFrame does not include"
                " 'generator_llm' and 'generator_llm' is not provided."
            )

        retrieved_contexts = workflow.get_retrieved_context(query)
        response = workflow.get_response(query)

        sample = RAGSample(
            user_input=query,
            retrieved_contexts=retrieved_contexts,
            response=response,
            reference=reference,
            generation_prompt=generation_prompt,
            generator_llm=generator_llm,
        )
        samples.append(sample)
    eval_dataset = RAGEvaluationDataset(samples)
    return eval_dataset


def auto_optimize(workflow: BaseNDRagWorkflow, n_trials: int):
    direction = "maximize" if workflow.objective_maximize else "minimize"
    study = optuna.create_study(
        study_name=workflow.job_name, direction=direction
    )
    study.optimize(workflow._outer_objective, n_trials=n_trials)
    workflow._set_param_values(study.best_params)
    return {"best_params": study.best_params, "trials": study.trials}


def _map_to_ragas_samples(
    dataset: RAGEvaluationDataset,
) -> Tuple[EvaluationDataset, pd.DataFrame]:
    ragas_samples = []
    extra_columns = {
        "generation_prompt": [],
    }
    for sample in dataset:
        ragas_sample = SingleTurnSample(
            user_input=sample.user_input,
            retrieved_contexts=sample.retrieved_contexts,
            reference_contexts=sample.reference_contexts,
            response=sample.response,
            multi_responses=sample.multi_responses,
            reference=sample.reference,
            rubrics=sample.rubrics,
        )
        ragas_samples.append(ragas_sample)
        extra_columns["generation_prompt"].append(sample.generation_prompt)

    extra_columns_df = pd.DataFrame.from_dict(extra_columns)
    return EvaluationDataset(ragas_samples), extra_columns_df


def _evaluate_dataset(
    generator_llm: LLMConfig,
    dataset: EvaluationDataset,
    metrics: Optional[Sequence[Metric]] = None,
    llm: Optional[Union[BaseRagasLLM, LangchainLLM]] = None,
    embeddings: Optional[
        Union[BaseRagasEmbeddings, LangchainEmbeddings]
    ] = None,
    callbacks: Callbacks = None,
    in_ci: bool = False,
    run_config: RunConfig = RunConfig(),
    token_usage_parser: Optional[TokenUsageParser] = None,
    raise_exceptions: bool = False,
    column_map: Optional[Dict[str, str]] = None,
    show_progress: bool = True,
    batch_size: Optional[int] = None,
) -> pd.DataFrame:
    LOGGER.info(f"Evaluating generations from {str(generator_llm)}")

    result = ragas_evaluate(
        dataset,
        metrics,
        llm,
        embeddings,
        callbacks,
        in_ci,
        run_config,
        token_usage_parser,
        raise_exceptions,
        column_map,
        show_progress,
        batch_size,
    )
    return result.to_pandas()


def _generate_rag_eval_dataset(
    generator_llm: LLMConfig,
    dataset: RAGEvaluationDataset,
    temperature: float = _DEFAULT_GENERATION_TEMPERATURE,
) -> RAGEvaluationDataset:
    LOGGER.info(f"Generating responses from {str(generator_llm)}")

    llm = get_llm(generator_llm)
    temperature = generator_llm.kwargs.get("temperature", temperature)

    eval_samples = []
    for sample in tqdm(dataset):
        response = llm.generate_text(
            StringPromptValue(text=sample.generation_prompt),
            temperature=temperature,
        )
        eval_sample = RAGSample(
            user_input=sample.user_input,
            retrieved_contexts=sample.retrieved_contexts,
            reference_contexts=sample.reference_contexts,
            response=response.generations[0][0].text,
            multi_responses=sample.multi_responses,
            reference=sample.reference,
            rubrics=sample.rubrics,
            generation_prompt=sample.generation_prompt,
            generator_llm=str(generator_llm),
        )
        eval_samples.append(eval_sample)
    return RAGEvaluationDataset(eval_samples)


@track_was_completed
def evaluate(
    dataset: RAGEvaluationDataset,
    metrics: Optional[Sequence[Metric]] = None,
    llm: Optional[Union[BaseRagasLLM, LangchainLLM]] = None,
    embeddings: Optional[
        Union[BaseRagasEmbeddings, LangchainEmbeddings]
    ] = None,
    callbacks: Callbacks = None,
    in_ci: bool = False,
    run_config: RunConfig = RunConfig(),
    token_usage_parser: Optional[TokenUsageParser] = None,
    raise_exceptions: bool = False,
    column_map: Optional[Dict[str, str]] = None,
    show_progress: bool = True,
    batch_size: Optional[int] = None,
    generator_llms: List[LLMConfig] = [],
) -> Dict[str, pd.DataFrame]:
    dataset_llm_str = dataset[0].generator_llm
    dataset_llm_config = LLMConfig.from_string(dataset_llm_str)

    if dataset_llm_config not in generator_llms:
        generator_llms.append(dataset_llm_config)

    ragas_dataset, extra_columns = _map_to_ragas_samples(dataset)

    dataset_results = _evaluate_dataset(
        dataset_llm_config,
        ragas_dataset,
        metrics,
        llm,
        embeddings,
        callbacks,
        in_ci,
        run_config,
        token_usage_parser,
        raise_exceptions,
        column_map,
        show_progress,
        batch_size,
    )

    evaluation_results = {
        str(dataset_llm_config): pd.concat(
            [dataset_results, extra_columns], axis=1
        )
    }

    for llm_config in generator_llms:
        if str(llm_config) in evaluation_results:
            continue

        llm_dataset = _generate_rag_eval_dataset(llm_config, dataset)
        ragas_dataset, extra_columns = _map_to_ragas_samples(llm_dataset)
        dataset_results = _evaluate_dataset(
            llm_config,
            ragas_dataset,
            metrics,
            llm,
            embeddings,
            callbacks,
            in_ci,
            run_config,
            token_usage_parser,
            raise_exceptions,
            column_map,
            show_progress,
            batch_size,
        )
        evaluation_results[str(llm_config)] = pd.concat(
            [dataset_results, extra_columns], axis=1
        )
    return evaluation_results



import importlib.util
import logging
import tiktoken
from typing import Dict, cast, List, Optional, Union

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def _module_check(module_name: str, attr_name: str, provider_name: str = None):
    parent_module_name = module_name.split(".")[0]

    if provider_name:
        provider_title = provider_name.title()
        if provider_title == "Openai":
            provider_title = "OpenAI"
        err_msg = (
            f"`notdiamond` requires `{parent_module_name}` to generate responses from {provider_title} models. "
            "Please install via `pip install notdiamond[create]`."
        )
    else:
        err_msg = (
            f"`notdiamond` requires `{parent_module_name}` to generate responses. Please install via "
            "`pip install notdiamond[create]`."
        )

    try:
        module = importlib.import_module(module_name)
        cls = getattr(module, attr_name)
    except (ModuleNotFoundError, AttributeError) as err:
        raise ModuleNotFoundError(err_msg) from err
    return cls


def convert_tool_to_openai_function(tool):
    convert_python_function_to_openai_function = None
    BaseTool = None
    format_tool_to_openai_function = None
    try:
        convert_python_function_to_openai_function = _module_check(
            "langchain_core.utils.function_calling",
            "convert_python_function_to_openai_function",
        )
        BaseTool = _module_check("langchain_core.tools", "BaseTool")
        format_tool_to_openai_function = _module_check(
            "langchain_core.utils.function_calling",
            "format_tool_to_openai_function",
        )
    except (ModuleNotFoundError, ImportError) as ierr:
        LOGGER.info(
            "Could not import function calling and tool helpers from langchain_core."
            f"{ierr}"
        )

    if isinstance(tool, dict) and all(
        k in tool for k in ("name", "description", "parameters")
    ):
        return tool
    # a JSON schema with title and description
    elif isinstance(tool, dict) and all(
        k in tool for k in ("title", "description", "properties")
    ):
        tool = tool.copy()
        return {
            "name": tool.pop("title"),
            "description": tool.pop("description"),
            "parameters": tool,
        }
    elif isinstance(tool, BaseTool):
        if not format_tool_to_openai_function:
            raise ValueError(
                "Provided tool is a langchain BaseTool, but langchain is not available to import. "
                "Please run pip install notdiamond[create] to install langchain."
            )
        return cast(Dict, format_tool_to_openai_function(tool))
    elif callable(tool):
        if not convert_python_function_to_openai_function:
            raise ValueError(
                "Provided tool is a Python function, but langchain is not available to import. "
                "Tool calling of functions is only available with langchain. Please run pip install notdiamond[create] "
                "to install langchain."
            )
        return cast(Dict, convert_python_function_to_openai_function(tool))

    else:
        raise ValueError(
            f"Unsupported function\n\n{tool}\n\nFunctions must be passed in"
            " as Dict, pydantic.BaseModel, or Callable. If they're a dict they must"
            " either be in OpenAI function format or valid JSON schema with top-level"
            " 'title' and 'description' keys."
        )


def _default_headers(
    notdiamond_api_key: str, user_agent: str
) -> Dict[str, str]:
    return {
        "content-type": "application/json",
        "Authorization": f"Bearer {notdiamond_api_key}",
        "User-Agent": user_agent,
    }


def token_counter(
    model: str = "",
    text: Optional[str] = None,
    messages: Optional[List[Dict[str, str]]] = None,
) -> int:
    """
    Count the number of tokens in a given text or messages using a specified model.
    
    Args:
        model (str): The name of the model to use for tokenization.
        text (Optional[str]): The raw text string to count tokens for.
        messages (Optional[List[Dict[str, str]]]): List of messages with "role" and "content" keys.
        
    Returns:
        int: The number of tokens in the text or messages.
    """
    if text is None and messages is None:
        raise ValueError("Either text or messages must be provided")
    
    if text is None:
        text = ""
        for message in messages:
            if message.get("content", None) is not None:
                content = message.get("content")
                if isinstance(content, str):
                    text += content
    
    # Use tiktoken for OpenAI models
    try:
        if "gpt-4" in model or "gpt-3.5" in model:
            encoding = tiktoken.encoding_for_model(model)
        else:
            # Default to cl100k_base for other models
            encoding = tiktoken.get_encoding("cl100k_base")
        
        return len(encoding.encode(text))
    except Exception as e:
        LOGGER.warning(f"Error using tiktoken: {e}. Using simple character count approximation.")
        # Fallback: rough approximation (4 chars ≈ 1 token)
        return len(text) // 4



import logging
from typing import Any

from notdiamond._utils import _module_check
from notdiamond.llms.config import LLMConfig

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def _nd_callback_handler_factory():
    try:
        BaseCallbackHandler = _module_check(
            "langchain_core.callbacks.base", "BaseCallbackHandler"
        )
    except (ModuleNotFoundError, ImportError) as ierr:
        log_msg = (
            "Attempted to instantiate NDLLMBaseCallbackHandler but langchain is not installed. ",
            f"Model callbacks not available. {ierr}",
        )
        LOGGER.warning(log_msg)
        return

    class _NDLLMBaseCallbackHandler(BaseCallbackHandler):
        """
        Base callback handler for NotDiamond LLMs.
        Accepts all of the langchain_core callbacks and adds new ones.
        """

        def on_model_select(
            self, model_provider: LLMConfig, model_name: str
        ) -> Any:
            """
            Called when a model is selected.
            """

        def on_latency_tracking(
            self,
            session_id: str,
            model_provider: LLMConfig,
            tokens_per_second: float,
        ):
            """
            Called when latency tracking is enabled.
            """

        def on_api_error(self, error_message: str):
            """
            Called when an NotDiamond API error occurs.
            """

    return _NDLLMBaseCallbackHandler


NDLLMBaseCallbackHandler = _nd_callback_handler_factory()



from dataclasses import dataclass
from typing import Dict, List, Union, overload

from ragas import MultiTurnSample, SingleTurnSample
from ragas.dataset_schema import RagasDataset


class RAGSample(SingleTurnSample):
    """
    Represents RAG evaluation samples.

    Attributes:
        user_input (Optional[str]): The input query from the user.
        retrieved_contexts (Optional[List[str]]): List of contexts retrieved for the query.
        reference_contexts (Optional[List[str]]): List of reference contexts for the query.
        response (Optional[str]): The generated response for the query.
        generation_prompt (str): The input prompt to the generator LLM.
        generator_llm (str): The LLM used to generate the response.
        multi_responses (Optional[List[str]]): List of multiple responses generated for the query.
        reference (Optional[str]): The reference answer for the query.
        rubric (Optional[Dict[str, str]]): Evaluation rubric for the sample.
    """

    generation_prompt: str
    generator_llm: str


@dataclass
class RAGEvaluationDataset(RagasDataset[RAGSample]):
    """
    Represents a dataset of RAG evaluation samples.

    Attributes:
        samples (List[BaseSample]): A list of evaluation samples.

    Methods:
        validate_samples(samples): Validates that all samples are of the same type.
        get_sample_type(): Returns the type of the samples in the dataset.
        to_hf_dataset(): Converts the dataset to a Hugging Face Dataset.
        to_pandas(): Converts the dataset to a pandas DataFrame.
        features(): Returns the features of the samples.
        from_list(mapping): Creates an EvaluationDataset from a list of dictionaries.
        from_dict(mapping): Creates an EvaluationDataset from a dictionary.
        to_csv(path): Converts the dataset to a CSV file.
        to_jsonl(path): Converts the dataset to a JSONL file.
        from_jsonl(path): Creates an EvaluationDataset from a JSONL file.
    """

    @overload
    def __getitem__(self, idx: int) -> RAGSample:
        ...

    @overload
    def __getitem__(self, idx: slice) -> "RAGEvaluationDataset":
        ...

    def __getitem__(
        self, idx: Union[int, slice]
    ) -> Union[RAGSample, "RAGEvaluationDataset"]:
        if isinstance(idx, int):
            return self.samples[idx]
        elif isinstance(idx, slice):
            return type(self)(samples=self.samples[idx])
        else:
            raise TypeError("Index must be int or slice")

    def is_multi_turn(self) -> bool:
        return False

    def to_list(self) -> List[Dict]:
        rows = [sample.to_dict() for sample in self.samples]
        return rows

    @classmethod
    def from_list(cls, data: List[Dict]):
        samples = []
        if all(
            "user_input" in item and isinstance(data[0]["user_input"], list)
            for item in data
        ):
            samples.extend(MultiTurnSample(**sample) for sample in data)
        else:
            samples.extend(SingleTurnSample(**sample) for sample in data)
        return cls(samples=samples)

    def __repr__(self) -> str:
        return f"RAGEvaluationDataset(features={self.features()}, len={len(self.samples)})"



"""
Private module for retry and fallback logic. Added to notdiamond.toolkit to avoid
introducing library dependencies on OpenAI into the core SDK.
"""
import asyncio
import logging
import random
import time
from functools import wraps
from typing import Any, Dict, List, Optional, Union

from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI

from notdiamond import NotDiamond
from notdiamond.llms.config import LLMConfig

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

ClientType = Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI]
ModelType = Union[Dict[str, float], List[str]]
OpenAIMessagesType = List[Dict[str, str]]


class _RetryWrapperException(Exception):
    failed_models: List[str]
    failed_exception: Exception

    def __init__(self, failed_models: List[str], failed_exception: Exception):
        self.failed_models = failed_models
        self.failed_exception = failed_exception
        super().__init__(
            f"Failed to invoke {failed_models}: {failed_exception}"
        )


class _CumulativeModelSelectionWeights:
    """
    Helper class for load balancing between models. Overloads `__getitem__` to
    return a model based on a random float.
    """

    def __init__(self, models: Dict[str, float]):
        """
        Args:
            models: Dict[str, float]
                A dictionary mapping model names to their selection weights. The weights will be normalized
                to sum to 1.0 and used for weighted random selection of models.
        """
        sorted_models = sorted(models.items(), key=lambda x: x[1])
        norm_sum = sum(weight for _, weight in sorted_models)
        if norm_sum == 0.0:
            raise ValueError(
                "All model weights are zero. Do not need to load balance."
            )
        self._cumulative_weights = []
        self._model_ordering = []
        previous_weight = 0.0
        for model, weight in sorted_models:
            weight_bin = (previous_weight, previous_weight + weight / norm_sum)
            self._cumulative_weights.append(weight_bin)
            self._model_ordering.append(model)
            previous_weight = weight_bin[1]

    def __getitem__(self, value: float) -> LLMConfig:
        for bin_idx, weight_bin in enumerate(self._cumulative_weights):
            if (
                weight_bin[0] <= value < weight_bin[1]
                or value == weight_bin[1] == 1.0
            ):
                return self._model_ordering[bin_idx]
        raise KeyError(
            f"No model found for draw {value} (model weights: {list(zip(self._model_ordering, self._cumulative_weights))})"
        )


class _BaseRetryWrapper:
    """
    Wrapper class for an individual OpenAI client which adds retry and fallback logic.
    """

    def __init__(
        self,
        client: ClientType,
        models: ModelType,
        max_retries: Union[int, Dict[str, int]] = 1,
        timeout: Union[float, Dict[str, float]] = 60.0,
        model_messages: OpenAIMessagesType = {},
        api_key: Union[str, None] = None,
        backoff: Union[float, Dict[str, float]] = 2.0,
    ):
        """
        Args:
            client: OpenAI, AsyncOpenAI, AzureOpenAI, or AsyncAzureOpenAI
                The client to wrap.
            models: Dict[str, float] | List[str]
                The models to invoke.
            max_retries: int | Dict[str, int]
                The maximum number of retries. Configured globally or per-model.
            timeout: float | Dict[str, float]
                The timeout for each model. Configured globally or per-model.
            model_messages: OpenAIMessagesType
                The messages to send to each model. Prepended to any messages passed to the `create` method.
            api_key: str | None
                Not Diamond API key to use for logging. Currently unused.
            backoff: float | Dict[str, float]
                Exponential backoff factor per each retry. Can be configured globally or per model.
        """
        self._client = client

        # track the default create method - we'll need to reference it when overriding create
        # also allows us to spy during tests later
        self._default_create = self._client.chat.completions.create

        # strip provider from model name
        if isinstance(models, list):
            models = [m for m in models if self._model_client_match(m)]
            # models is a list - preserve order
            self._models = [_m.split("/")[-1] for _m in models]
        else:
            # models is a load-balanced dict - order models by desc. weight
            self._models = {
                m.split("/")[-1]: w
                for m, w in models.items()
                if self._model_client_match(m)
            }

        self._max_retries = (
            {
                m.split("/")[-1]: t
                for m, t in max_retries.items()
                if self._model_client_match(m)
            }
            if isinstance(max_retries, dict)
            else max_retries
        )
        self._timeout = (
            {
                m.split("/")[-1]: t
                for m, t in timeout.items()
                if self._model_client_match(m)
            }
            if isinstance(timeout, dict)
            else timeout
        )
        self._model_messages = (
            {
                m.split("/")[-1]: msgs
                for m, msgs in model_messages.items()
                if self._model_client_match(m)
            }
            if model_messages
            else {}
        )
        self._backoff = (
            {
                m.split("/")[-1]: b
                for m, b in backoff.items()
                if self._model_client_match(m)
            }
            if isinstance(backoff, dict)
            else backoff
        )

        self._api_key = api_key
        self._nd_client = None
        if self._api_key:
            self._nd_client = NotDiamond(api_key=self._api_key)

        # track RetryManager to assist with model selection and load balancing when overriding
        # create or stream calls
        self.manager: Optional[RetryManager] = None

    def get_provider(self) -> str:
        """
        Get the provider name for this client - "azure" or "openai".
        """
        # platform child clients should precede the model-building clients (eg. Azure before OpenAI)
        # since the former usually inherit from the latter.
        if isinstance(self._client, (AsyncAzureOpenAI, AzureOpenAI)):
            return "azure"
        elif isinstance(self._client, (AsyncOpenAI, OpenAI)):
            return "openai"

    def _model_client_match(self, target_model: str) -> bool:
        if isinstance(self._client, (AsyncAzureOpenAI, AzureOpenAI)):
            return target_model.split("/")[0] == "azure"
        elif isinstance(self._client, (AsyncOpenAI, OpenAI)):
            return target_model.split("/")[0] == "openai"
        else:
            raise ValueError(
                f"No client match found for model {target_model}. Client type: {type(self._client)}"
            )

    def get_timeout(self, target_model: Optional[str] = None) -> float:
        """
        Get the configured timeout (if per-model, for the target model).
        """
        out = self._timeout
        if isinstance(self._timeout, dict):
            if not target_model:
                raise ValueError(
                    "target_model must be provided if timeout is a dict"
                )
            out = self._timeout.get(target_model)
        return out

    def get_max_retries(self, target_model: Optional[str] = None) -> int:
        """
        Get the configured max retries (if per-model, for the target model).
        """
        out = self._max_retries
        if isinstance(self._max_retries, dict):
            if not target_model:
                raise ValueError(
                    "target_model must be provided if max_retries is a dict"
                )
            out = self._max_retries.get(target_model)
        return out

    def get_backoff(self, target_model: Optional[str] = None) -> float:
        """
        Get the configured backoff (if per-model, for the target model).
        """
        out = self._backoff
        if isinstance(self._backoff, dict):
            if not target_model:
                raise ValueError(
                    "target_model must be provided if backoff is a dict"
                )
            out = self._backoff.get(target_model)
        return out

    def _update_model_kwargs(
        self,
        kwargs: Dict[str, Any],
        target_model: str,
        user_messages: Optional[List[Dict[str, Any]]] = [],
    ) -> Dict[str, Any]:
        kwargs["model"] = target_model
        kwargs["timeout"] = self.get_timeout(target_model)
        kwargs["messages"] = (
            self._model_messages[target_model] + user_messages
            if self._model_messages.get(target_model)
            else user_messages
        )
        return kwargs

    def _get_fallback_model(self, failed_models: List[str] = []) -> str:
        """
        After failing to invoke current_model, use the user's model fallback list to choose the
        next invocation model.
        """
        # todo [a9] remove this so RetryManager handles selection
        models = [m for m in self._models if m not in failed_models]

        if not models:
            return
        return models[0]

    def _retry_decorator(self, func):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            target_model = kwargs["model"]
            base_messages = kwargs.get("messages", [])
            kwargs = self._update_model_kwargs(
                kwargs, target_model, user_messages=base_messages
            )

            attempt = 0
            while attempt < self.get_max_retries(target_model):
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    LOGGER.warning(f"Attempt {attempt + 1} failed: {str(e)}")

                    if attempt == self.get_max_retries(target_model) - 1:
                        # throw exception - will invoke next model outside of wrapper
                        raise _RetryWrapperException([target_model], e)

                attempt += 1
                await asyncio.sleep(self.get_backoff(target_model) ** attempt)

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            target_model = kwargs["model"]
            base_messages = kwargs.get("messages", [])
            kwargs = self._update_model_kwargs(
                kwargs, target_model, user_messages=base_messages
            )

            attempt = 0
            while attempt < self.get_max_retries(target_model):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    LOGGER.warning(f"Attempt {attempt + 1} failed: {str(e)}")

                    if attempt == self.get_max_retries(target_model) - 1:
                        raise _RetryWrapperException([target_model], e)

                attempt += 1
                time.sleep(self.get_backoff(target_model) ** attempt)

        if isinstance(self, AsyncRetryWrapper):
            return async_wrapper
        return sync_wrapper

    def _update_failed_models(
        self,
        target_model: str,
        failed_models: List[str],
        all_failed_models: List[str],
    ) -> List[str]:
        target_provider = target_model.split("/")[0]
        all_failed_models.extend(
            ["/".join([target_provider, m]) for m in failed_models]
        )

    def __getattr__(self, name):
        return getattr(self._client, name)

    def _log_to_nd(self, *args, **kwargs):
        if not self._api_key:
            LOGGER.warning(
                "No API key provided, skipping logging to Not Diamond."
            )

        LOGGER.info("Logging inference metadata to Not Diamond.")
        raise NotImplementedError("Not Diamond logging not implemented.")


class _BaseCompletions:
    def __init__(self, parent: "_BaseRetryWrapper", manager: "RetryManager"):
        self.parent = parent
        self.manager = manager

    @property
    def completions(self):
        return self

    def _handle_retry_exception(
        self,
        e: _RetryWrapperException,
        target_model: str,
        all_failed_models: list,
    ) -> tuple:
        """Common exception handling logic"""
        LOGGER.exception(e)
        self.parent._update_failed_models(
            target_model, e.failed_models, all_failed_models
        )
        target_model = self.manager.get_next_model(all_failed_models)
        if not target_model:
            raise e.failed_exception
        target_wrapper = self.manager.get_wrapper(target_model)
        return target_model, target_wrapper, target_wrapper._default_create


class RetryWrapper(_BaseRetryWrapper):
    """
    Wrapper for OpenAI clients which adds retry and fallback logic. This method
    patches the `chat` property of the client to return a wrapped ChatCompletions
    object.

    If you need to patch the `create` method (eg. during unit testing), you can target
    the _default_create method of this class.
    """

    class ChatCompletions(_BaseCompletions):
        def create(self, *args, **kwargs):
            target_model = kwargs["model"]
            target_model = f"{self.parent.get_provider()}/{target_model}"
            target_wrapper = self.manager.get_wrapper(target_model)

            all_failed_models = []
            _create_fn = self.parent._default_create
            while True:
                wrapped = target_wrapper._retry_decorator(_create_fn)
                kwargs["model"] = target_model.split("/")[-1]

                try:
                    return wrapped(*args, **kwargs)
                except _RetryWrapperException as e:
                    (
                        target_model,
                        target_wrapper,
                        _create_fn,
                    ) = self._handle_retry_exception(
                        e, target_model, all_failed_models
                    )

    @property
    def chat(self):
        if not isinstance(self._client, OpenAI):
            return getattr(self._client, "chat")

        return self.ChatCompletions(self, self.manager)


class AsyncRetryWrapper(_BaseRetryWrapper):
    """
    Async wrapper for OpenAI clients which adds retry and fallback logic. This method
    patches the `chat` property of the client to return a wrapped AsyncCompletions
    object.

    If you need to patch the `create` method (eg. during unit testing), you can target
    the _default_create method of this class.
    """

    class AsyncCompletions(_BaseCompletions):
        async def create(self, *args, **kwargs):
            target_model = kwargs["model"]
            target_model = f"{self.parent.get_provider()}/{target_model}"
            target_wrapper = self.manager.get_wrapper(target_model)

            all_failed_models = []
            _create_fn = self.parent._default_create

            while True:
                wrapped = target_wrapper._retry_decorator(_create_fn)
                kwargs["model"] = target_model.split("/")[-1]

                try:
                    return await wrapped(*args, **kwargs)
                except _RetryWrapperException as e:
                    (
                        target_model,
                        target_wrapper,
                        _create_fn,
                    ) = self._handle_retry_exception(
                        e, target_model, all_failed_models
                    )

    @property
    def chat(self):
        if not isinstance(self._client, AsyncOpenAI):
            return getattr(self._client, "chat")

        return self.AsyncCompletions(self, self.manager)


class RetryManager:
    """
    RetryManager handles model selection and load balancing when using `notdiamond.init`.
    """

    def __init__(
        self,
        models: Union[Dict[str, float], List[str]],
        wrapped_clients: List[Union[RetryWrapper, AsyncRetryWrapper]],
    ):
        """
        Args:
            models: Dict[str, float] | List[str]
                The models to invoke.
            wrapped_clients: List[RetryWrapper | AsyncRetryWrapper]
                Clients wrapped by this manager.
        """
        self._wrappers = wrapped_clients
        self._models = models

        self._model_weights = (
            _CumulativeModelSelectionWeights(self._models)
            if isinstance(self._models, dict)
            else None
        )

        # Attach the manager to each wrapper and patch chat behavior
        # Without this hack we cannot wrap the create method successfully
        for wrapper in self._wrappers:
            wrapper.manager = self
            wrapper._client.chat = wrapper.chat

        self._model_to_wrapper = {
            model: self._get_model_wrapper(model) for model in self._models
        }

    def _get_model_wrapper(
        self, model: str
    ) -> Union[RetryWrapper, AsyncRetryWrapper]:
        target_wrapper = None
        try:
            if "azure" in model:
                instance_check = (AzureOpenAI, AsyncAzureOpenAI)
            elif "openai" in model:
                instance_check = (OpenAI, AsyncOpenAI)
            else:
                raise ValueError(
                    f"No wrapper found for model {model}. It may not currently be supported."
                )

            target_wrapper = [
                wrapper
                for wrapper in self._wrappers
                if isinstance(wrapper._client, instance_check)
            ][0]
        except IndexError:
            raise ValueError(
                f"No wrapped client found for model {model} among {[w._client for w in self._wrappers]}."
            )

        return target_wrapper

    def get_next_model(self, failed_models: List[str]) -> str:
        """
        Select the next model to invoke. Omit all models that have previously failed.
        """
        if not self._model_weights:
            remaining_models = [
                m for m in self._models if m not in failed_models
            ]
            if not remaining_models:
                return
            return remaining_models[0]

        try:
            new_weights = _CumulativeModelSelectionWeights(
                {
                    m: w
                    for m, w in self._models.items()
                    if m not in failed_models
                }
            )
        except ValueError:
            # model weights are all zero - do not need to load balance
            return
        return new_weights[random.random()]

    def _select_target_model(self, kwargs: Dict[str, Any]) -> str:
        target_model = kwargs.get("model", None)
        if target_model == "notdiamond":
            return self._model_weights[random.random()]
        return target_model

    def get_wrapper(
        self, model: str
    ) -> Union[RetryWrapper, AsyncRetryWrapper]:
        try:
            return self._model_to_wrapper[model]
        except KeyError:
            raise ValueError(f"No wrapper found for model {model}.")
